Baggins = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceDB-2.0", "AceEvent-2.0", "AceDebug-2.0", "FuBarPlugin-2.0", "AceHook-2.1")
local pt = AceLibrary("PeriodicTable-2.0")
local L = AceLibrary("AceLocale-2.2"):new("Baggins")
local tablet = AceLibrary("Tablet-2.0")
local dewdrop = AceLibrary("Dewdrop-2.0")
local gratuity = AceLibrary("Gratuity-2.0")

Baggins.hasIcon = "Interface\\Icons\\INV_Jewelry_Ring_03"
Baggins.cannotDetachTooltip = true
Baggins.clickableTooltip = true
Baggins.independentProfile = true
Baggins.hideWithoutStandby = true

BINDING_HEADER_BAGGINS = L["Baggins"]
BINDING_NAME_BAGGINS_TOGGLEALL = L["Toggle All Bags"]

local categorycache = {}
local useditems = {}
local slotcache = {}
local itemcounts = {}

local bankuseditems = {}
local bankcategorycache = {}

local currentbag
local currentsection
local currentcategory

local ptsets

local catsorttable = {}

local BagNames = {
	[0] = L["Backpack"],
	[1] = L["Bag1"],
	[2] = L["Bag2"],
	[3] = L["Bag3"],
	[4] = L["Bag4"],
	[-1] = L["Bank Frame"],
	[5] = L["Bank Bag1"],
	[6] = L["Bank Bag2"],
	[7] = L["Bank Bag3"],
	[8] = L["Bank Bag4"],
	[9] = L["Bank Bag5"],
	[10] = L["Bank Bag6"],
	[11] = L["Bank Bag7"],	
	[-2] = L["KeyRing"],
}

local BagTypes = { 
	--normal bags
	[0] = 1,
	[1] = 1,
	[2] = 1,
	[3] = 1,
	[4] = 1,
	--bank bags
	[-1] = 2,
	[5] = 2,
	[6] = 2,
	[7] = 2,
	[8] = 2,
	[9] = 2,
	[10] = 2,
	[11] = 2,
	--keyring
	[-2] = 3,		
}

local QualityNames = {
	[0] = "Poor",
	[1] = "Common",
	[2] = "Uncommon",
	[3] = "Rare",
	[4] = "Epic",
	[5] = "Legendary",
	[6] = "Artifact",
}

local ItemTypes = { 
	["Armor"] = { "Cloth", "Idols", "Leather", "Librams", "Mail", "Miscellaneous", "Shields", "Totems", "Plate"},
	["Consumable"] = { "Consumable" },
	["Container"]  = {"Bag", "Enchanting Bag", "Engineering Bag", "Herb Bag", "Soul Bag" },
	["Key"] = { "Key" },
	["Miscellaneous"] = { "Junk" },
	["Reagent"]  = { "Reagent" },
	["Recipe"]  = { "Alchemy","Blacksmithing","Book","Cooking","Enchanting","Engineering","First Aid","Leatherworking","Tailoring" },
	["Projectile"]  = {"Arrow","Bullet"},
	["Quest"]  = { "Quest" },
	["Quiver"]  = { "Ammo Pouch","Quiver" },
	["Trade Goods"]  = { "Devices","Explosives","Parts","Trade Goods" },
	["Weapon"]  = { "Bows","Crossbows","Daggers","Guns","Fishing Pole","Fist Weapons","Miscellaneous","One-Handed Axes","One-Handed Maces","One-Handed Swords","Polearms","Staves","Thrown","Two-Handed Axes","Two-Handed Maces","Two-Handed Swords","Wands"},
}

local EquipLocs = {
	"INVTYPE_AMMO",
	"INVTYPE_HEAD",
	"INVTYPE_NECK",
	"INVTYPE_SHOULDER",
	"INVTYPE_BODY", 
	"INVTYPE_CHEST",
	"INVTYPE_ROBE",
	"INVTYPE_WAIST",
	"INVTYPE_LEGS",
	"INVTYPE_FEET",
	"INVTYPE_WRIST",
	"INVTYPE_HAND",
	"INVTYPE_FINGER",
	"INVTYPE_TRINKET",
	"INVTYPE_CLOAK",
	"INVTYPE_WEAPON",
	"INVTYPE_SHIELD",
	"INVTYPE_2HWEAPON",
	"INVTYPE_WEAPONMAINHAND",
	"INVTYPE_WEAPONOFFHAND",
	"INVTYPE_HOLDABLE",
	"INVTYPE_RANGED",
	"INVTYPE_THROWN",
	"INVTYPE_RANGEDRIGHT",
	"INVTYPE_RELIC",
	"INVTYPE_TABARD",
	"INVTYPE_BAG"
}
--removes any fields not used by the current rule type and sets up defaults if needed
function Baggins:CleanRule(rule)
	if rule.type ~= "ItemType" then
		rule.itype = nil
		rule.isubtype = nil
	else
		rule.itype = "Miscellaneous"
	end
	if rule.type ~= "ContainerType" then
		rule.ctype = nil
	else
		--rule.ctype = "
	end
	if rule.type ~= "Bag" then
		rule.bagid = nil
		rule.noempty = nil
	else
		rule.bagid = 0
	end
	if rule.type ~= "PTSet" then
		rule.setname = nil
	end
	if rule.type ~= "EquipLoc" then
		rule.equiploc = nil
	else
		rule.equiploc = EquipLocs[1]
	end
	if rule.type ~= "Quality" then
		rule.quality = nil
		rule.comp = nil
	else
		rule.quality = 1
		rule.comp = "=="
	end
	if rule.type ~= "ItemName" then
		rule.match = nil
	end
end

local RuleTypes = {
	ItemType = {
		DisplayName = L["Item Type"],
		Description = L["Filter by Item type and sub-type as returned by GetItemInfo"],
		Matches = function(bag,slot,rule) 
			if not (rule.itype or rule.isubtype) then return end
			local link = GetContainerItemLink(bag, slot)
			if link then
				local _, _, _, _, _, Type, SubType = GetItemInfo(link)
				if Type and SubType then
					return Type == L[rule.itype] and (rule.isubtype == nil or L[rule.isubtype] == SubType)	
				end
			end
		end,
		GetName = function(rule) 
			local type, subtype
			if rule.itype then
				type = L[rule.itype]
			else
				type = "*"
			end
			if rule.isubtype then
				subtype = L[rule.isubtype]
			else
				subtype = "*"
			end
			return L["ItemType - "]..type..":"..subtype
		end,
		DewDropOptions = function(rule, level, value) 
			if level == 1 then
				dewdrop:AddLine('text', L["Item Type Options"], 'isTitle', true)
				dewdrop:AddLine('text', L["Item Type"], 'hasArrow', true,'value',"ItemType")
				dewdrop:AddLine('text', L["Item Subtype"], 'hasArrow', true,'value',"ItemSubtype")

			elseif level == 2 and value == "ItemType" then	
				for k, v in pairs(ItemTypes) do
					dewdrop:AddLine('text', L[k], "checked", rule.itype == k,"func",function(k) rule.itype = k rule.isubtype = nil Baggins:OnRuleChanged() end,"arg1",k)
				end
			elseif level == 2 and value == "ItemSubtype" then
				dewdrop:AddLine('text', L["All"], "checked", rule.isubtype == nil,"func",function() rule.isubtype = nil Baggins:OnRuleChanged() end)
				dewdrop:AddLine()
				if rule.itype and ItemTypes[rule.itype] then
					for k, v in ipairs(ItemTypes[rule.itype]) do
						dewdrop:AddLine('text', L[v], "checked", rule.isubtype == v,"func",function(v) rule.isubtype = v Baggins:OnRuleChanged() end,"arg1",v)
					end
				end
			end
		end,
	},
	ContainerType = {
		DisplayName = L["Container Type"],
		Description = L["Filter by the type of container the item is in."],	
		Matches = function(bag,slot,rule)
			if bag < 1 or bag > 4 then return end
			if not rule.ctype then return end
			local link = GetInventoryItemLink("player",19+bag)
			if link then
				local _, _, _, _, _, _, SubType = GetItemInfo(link)
				if SubType then
					return SubType == L[rule.ctype]
				end
			end
		end,
		GetName = function(rule) 
			local ctype
			if rule.ctype then
				ctype = L[rule.ctype]
			else
				ctype = L["None"]
			end
			return L["Container : "]..ctype
		end,
		DewDropOptions = function(rule, level, value) 
			if level == 1 then
				dewdrop:AddLine('text', L["Container Type Options"], 'isTitle', true)
				dewdrop:AddLine('text', L["Container Type"], 'hasArrow', true,'value',"ContainerType")
	
			elseif level == 2 and value == "ContainerType" then
				dewdrop:AddLine('text', L["All"], "checked", rule.ctype == nil,"func",function() rule.ctype = nil Baggins:OnRuleChanged() end)
				dewdrop:AddLine()
				for k, v in ipairs(ItemTypes["Container"]) do
					dewdrop:AddLine('text', L[v], "checked", rule.ctype == v,"func",function(v) rule.ctype = v Baggins:OnRuleChanged() end,"arg1",v)
				end	
				for k, v in ipairs(ItemTypes["Quiver"]) do
					dewdrop:AddLine('text', L[v], "checked", rule.ctype == v,"func",function(v) rule.ctype = v Baggins:OnRuleChanged() end,"arg1",v)
				end		
			end
		end,
	},
	ItemID = {
		DisplayName = L["Item ID"],
		Description = L["Filter by ItemID, this can be a space delimited list or ids to match."],
		Matches = function(bag,slot,rule) 
			if not rule.ids then return end
			local link = GetContainerItemLink(bag, slot)
			if link then
				local itemid = string.match(link, "item:(%d+)")
				return rule.ids[tonumber(itemid)]
			end
		end,
		GetName = function(rule) 
			return L["ItemIDs "]
		end,
		DewDropOptions = function(rule, level, value) 
			if level == 1 then
				dewdrop:AddLine('text', L["ItemID Options"], 'isTitle', true)
				dewdrop:AddLine('text', L["Current IDs, click to remove"], 'isTitle', true)
				if rule.ids then
					for k, v in pairs(rule.ids) do
						local name = GetItemInfo(k)
						dewdrop:AddLine('text', k.." "..(name or ""),"func", function(id) rule.ids[id] = nil end, "arg1", k )
					end
				end
				dewdrop:AddLine('text', L["New"], 'hasArrow', true,'value',"NewID","hasEditBox",true,"editBoxText", "",
								"editBoxFunc", function(text) 
									local id = tonumber(text)
									if id ~= 0 then
										if not rule.ids then
											rule.ids = {}
										end
										rule.ids[id] = true 
									end
									Baggins:OnRuleChanged() 
								end )

			end
		end,
	},
	Bag = {
		DisplayName = L["Bag"],
		Description = L["Filter by the bag the item is in"],	
		Matches = function(bag,slot,rule) 
			if rule.noempty then
				local link = GetContainerItemLink(bag, slot)
				if link then
					return bag == rule.bagid 
				end
			else
				return bag == rule.bagid
			end
		end,
		GetName = function(rule) 
			return "Bag "..(BagNames[rule.bagid] or rule.bagid or "*none*")..((rule.noempty and " *NotEmpty*") or "")
		end,
		DewDropOptions = function(rule, level, value) 
			if level == 1 then
				dewdrop:AddLine('text', L["Bag Options"], 'isTitle', true)
				dewdrop:AddLine('text', L["Bag"], 'hasArrow', true,'value',"Bag")
				dewdrop:AddLine('text', L["Ignore Empty Slots"], "checked", rule.noempty ,"func",function() rule.noempty = not rule.noempty Baggins:OnRuleChanged() end)	
			elseif level == 2 and value == "Bag" then
				dewdrop:AddLine('text', L["Backpack"], "checked", rule.bagid == 0,"func",function() rule.bagid = 0 Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', L["Bag1"], "checked", rule.bagid == 1,"func",function() rule.bagid = 1 Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', L["Bag2"], "checked", rule.bagid == 2,"func",function() rule.bagid = 2 Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', L["Bag3"], "checked", rule.bagid == 3,"func",function() rule.bagid = 3 Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', L["Bag4"], "checked", rule.bagid == 4,"func",function() rule.bagid = 4 Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', L["KeyRing"], "checked", rule.bagid == -2,"func",function() rule.bagid = -2 Baggins:OnRuleChanged() end)
			end
		end,
	},
	PTSet = {
		DisplayName = L["PeriodicTable Set"],
		Description = L["Filter by PeriodicTable Set"],	
		Matches = function(bag,slot,rule) 
			if not rule.setname then return end
			local link = GetContainerItemLink(bag, slot)
			if link then
				local itemid = string.match(link, "item:(%d+)")
				itemid = tonumber(itemid)
				return pt:ItemInSet(itemid,rule.setname)
			end
		end,
		GetName = function(rule) 
			return "PTSet:"..(rule.setname or "*none*")
		end,
		DewDropOptions = function(rule, level, value, value_1, value_2) 
			if level == 1 then
				dewdrop:AddLine('text', L["Periodic Table Set Options"], 'isTitle', true)
				dewdrop:AddLine('text', L["Set"], 'hasArrow', true,'value',"SetName")	
			elseif (level == 2 and value == "SetName") or (level >= 3) then
				Baggins:PTSetDewdrop(rule, level-1, value, value_1)
			end
		end,
	},
	ItemName = {
		DisplayName = L["Item Name"],
		Description = L["Filter by Name or partial name"],	
		Matches = function(bag,slot,rule) 
			if not rule.match then return end
			local link = GetContainerItemLink(bag, slot)
			
			if link then
				local itemname = GetItemInfo(link)
				if itemname:lower():match(rule.match:lower()) then
					return true
				end
			end
		end,
		GetName = function(rule) 
			return L["Name: "]..(rule.match or "")
		end,
		DewDropOptions = function(rule, level, value) 
			if level == 1 then
				dewdrop:AddLine('text', L["Item Name Options"], 'isTitle', true)
				dewdrop:AddLine('text', L["String to Match"], 'hasArrow', true,"hasEditBox",true,"editBoxText", rule.match or "","editBoxFunc", function(text) rule.match = text Baggins:OnRuleChanged() end )	
			end
		end,
	},
	Empty = {
		DisplayName = L["Empty Slots"],
		Description = L["Empty bag slots"],
		Matches = function(bag,slot,rule) 
			if not (bag and slot) then return end
			if BagTypes[bag] == 3 then return end
			local link = GetContainerItemLink(bag, slot)
			if not link then 
				return true 
			end
		end,
		GetName = function(rule) 
			return L["Empty Slots"]
		end,
		DewDropOptions = function(rule, level, value) 
		end,
	},
	AmmoBag = {
		DisplayName = L["Ammo Bag"],
		Description = L["Items in an ammo pouch or quiver"],
		Matches = function(bag,slot,rule) 
			if bag < 1 or bag > 4 then return end
			local link = GetInventoryItemLink("player",19+bag)
			if link then
				local _, _, _, _, _, Type = GetItemInfo(link)
				if Type then
					return Type == L["Quiver"]
				end
			end
		end,
		GetName = function(rule) 
			return L["Ammo Bag Slots"]
		end,
		DewDropOptions = function(rule, level, value)
		end,
	},
	Quality = {
		DisplayName = L["Quality"],
		Description = L["Filter by Item Quality"],
		Matches = function(bag,slot,rule) 
			if not (rule.comp and rule.quality) then return end
			local link = GetContainerItemLink(bag, slot)
			if link then
				local _, _, Rarity = GetItemInfo(link)
				if Rarity then
					return ( rule.comp == "==" and Rarity == rule.quality ) or
					       ( rule.comp == "<=" and Rarity <= rule.quality ) or
					       ( rule.comp == ">=" and Rarity >= rule.quality )
				end
			end
		end,
		GetName = function(rule) 
			if rule.quality then
				local r,g,b,hex = GetItemQualityColor(rule.quality)
				qualname = hex..QualityNames[rule.quality]
			else
				qualname = "*none*"
			end
			return L["Quality"].." "..(rule.comp or "==").." "..qualname
		end,
		DewDropOptions = function(rule, level, value)
			if level == 1 then
				dewdrop:AddLine('text', L["Quality Options"], 'isTitle', true)
				dewdrop:AddLine('text', L["Quality"], 'hasArrow', true,'value',"Quality")
				dewdrop:AddLine('text', L["Comparison"], 'hasArrow', true, 'value', "QualComp")
		
			elseif level == 2 and value == "Quality" then
				for i = 0, 6 do
					local r,g,b = GetItemQualityColor(i)
					dewdrop:AddLine('text', QualityNames[i], "checked", rule.quality == i,"func",function() rule.quality = i Baggins:OnRuleChanged() end,
									'textR',r,'textG',g,'textB',b)
				end
			elseif level == 2 and value == "QualComp" then		
				dewdrop:AddLine('text', "==", "checked", rule.comp == "==","func",function() rule.comp = "==" Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', "<=", "checked", rule.comp == "<=","func",function() rule.comp = "<=" Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', ">=", "checked", rule.comp == ">=","func",function() rule.comp = ">=" Baggins:OnRuleChanged() end)
			end
		end,
	},
	EquipLoc = {
		DisplayName = L["Equip Location"],
		Description = L["Filter by Equip Location as returned by GetItemInfo"],
		Matches = function(bag,slot,rule) 
			if not rule.equiploc then return end
			local link = GetContainerItemLink(bag, slot)
			if link then
				local _, _, _, _, _, _, _, _, EquipLoc = GetItemInfo(link)
				if EquipLoc then
					return EquipLoc == rule.equiploc
				end
			end
		end,
		GetName = function(rule) 
			return "Equip Location: "..(getglobal(rule.equiploc) or rule.equiploc or "*None*")
		end,
		DewDropOptions = function(rule, level, value) 
			if level == 1 then
				dewdrop:AddLine('text', L["Equip Location Options"], 'isTitle', true)
				dewdrop:AddLine('text', L["Location"], 'hasArrow', true,'value',"Location")

			elseif level == 2 and value == "Location" then	
				for k, v in pairs(EquipLocs) do
					dewdrop:AddLine('text', getglobal(v) or v or "", "checked", rule.equiploc == v,"func",function(v) rule.equiploc = v Baggins:OnRuleChanged() end,"arg1",v)
				end
			end
		end,
	},
	Other = {
		DisplayName = L["Unfiltered Items"],
		Description = L["Matches all items that arent matched by any other bag, NOTE: this should be the only rule in a category, others will be ignored"],
		Matches = function(bag,slot,rule) 
			
			--local key = bag..":"..slot
			--return not useditems[key]
		end,
		GetName = function(rule) 
			return L["Unfiltered"]
		end,
		DewDropOptions = function(rule, level, value) 
		
		end,
	},
	Bind = {
		DisplayName = L["Bind"],
		Description = L["Filter based on if the item binds, or if it is already bound"],
		Matches = function(bag, slot, rule)
			if not rule.bindtype then return end
			if rule.bindtype == "Soulbound" then
				if bag == -1 then
					gratuity:SetInventoryItem("player", BankButtonIDToInvSlotID(slot))
				else
					gratuity:SetBagItem(bag,slot)
				end
				if gratuity:Find(ITEM_SOULBOUND,2,2,false,false,true) then
					return true
				end
			elseif rule.bindtype == "Unbound" then
				local link = GetContainerItemLink(bag,slot)
					if link then
					if bag == -1 then
						gratuity:SetInventoryItem("player", BankButtonIDToInvSlotID(slot))
					else
						gratuity:SetBagItem(bag,slot)
					end
					if not gratuity:Find(ITEM_SOULBOUND,2,2,false,false,true) then
						return true
					end
				end
			elseif rule.bindtype == "BoP" then
				local link = GetContainerItemLink(bag,slot)
				if link then
					gratuity:SetHyperlink(link)
					if gratuity:Find(ITEM_BIND_ON_PICKUP,2,2) then
						return true
					end
				end
			elseif rule.bindtype == "BoE" then
				local link = GetContainerItemLink(bag,slot)
				if link then
					gratuity:SetHyperlink(link)
					if gratuity:Find(ITEM_BIND_ON_EQUIP,2,2) then
						return true
					end
				end
			elseif rule.bindtype == "BoU" then
				local link = GetContainerItemLink(bag,slot,2,2)
				if link then
					gratuity:SetHyperlink(link)
					if gratuity:Find(ITEM_BIND_ON_USE,2,2) then
						return true
					end
				end
			end
		end,
		GetName = function(rule)
			if rule.bindtype == "Soulbound" then
				return ITEM_SOULBOUND
			elseif rule.bindtype == "Unbound" then
				return L["Unbound"]
			elseif rule.bindtype == "BoP" then
				return ITEM_BIND_ON_PICKUP
			elseif rule.bindtype == "BoE" then
				return ITEM_BIND_ON_EQUIP
			elseif rule.bindtype == "BoU" then
				return ITEM_BIND_ON_USE
			else
				return L["Bind *unset*"]
			end
		end,
		DewDropOptions = function(rule, level, value)
			if level == 1 then
				dewdrop:AddLine('text', L["Bind Options"], 'isTitle', true)
				dewdrop:AddLine('text', L["Bind Type"], 'hasArrow', true,'value',"BindType")

			elseif level == 2 and value == "BindType" then	
				dewdrop:AddLine('text', ITEM_SOULBOUND, "checked", rule.bindtype == "Soulbound","func",function() rule.bindtype = "Soulbound" Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', L["Unbound"], "checked", rule.bindtype == "Unbound","func",function() rule.bindtype = "Unbound" Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', ITEM_BIND_ON_PICKUP, "checked", rule.bindtype == "BoP","func",function() rule.bindtype = "BoP" Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', ITEM_BIND_ON_EQUIP, "checked", rule.bindtype == "BoE","func",function() rule.bindtype = "BoE" Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', ITEM_BIND_ON_USE, "checked", rule.bindtype == "BoU","func",function() rule.bindtype = "BoU" Baggins:OnRuleChanged() end)
			end
		end,
	},
	Tooltip = {
		DisplayName = L["Tooltip"],
		Description = L["Filter based on text contained in its tooltip"],
		Matches = function(bag, slot, rule)
			if not rule.text then return end
			local text = rule.text
			--if the text is the name of a global string then match against that
			if text:upper() == text then
				local gtext = getglobal(rule.text)
				if type(gtext) == "string" then
					text = gtext
				end
			end
			if bag == -1 then
				gratuity:SetInventoryItem("player", BankButtonIDToInvSlotID(slot))
			else
				gratuity:SetBagItem(bag,slot)
			end
			if gratuity:Find(text) then
				return true
			end
		end,
		GetName = function(rule)
			return L["Tooltip"]..": "..(rule.text or "")
		end,
		DewDropOptions = function(rule, level, value)
			if level == 1 then
				dewdrop:AddLine('text', L["Tooltip Options"], 'isTitle', true)
				dewdrop:AddLine('text', L["String to Match"], 'hasArrow', true,"hasEditBox",true,"editBoxText", rule.text or "","editBoxFunc", function(text) rule.text = text Baggins:OnRuleChanged() end )	
			end
		end,
	}
}

--deep copy of a table, will NOT handle tables as keys or circular references
local function deepCopy(to, from)
	for k in pairs(to) do
		to[k] = nil
	end

	for k, v in pairs(from) do
		if type(v) == "table" then
			to[k] = {}
			deepCopy(to[k], from[k])
		else
			to[k] = from[k]
		end
	end
end

function Baggins:OnInitialize() 
    --self:RegisterEvent("BAG_UPDATE","OnBagUpdate")
    self.bagframes = {}
	self.itemcache = {}
	self.colors = {
		black = {r=0,g=0,b=0,hex="|cff000000"},
		white = {r=1,g=1,b=1,hex="|cffffffff"},
		blue = {r=0,g=0.5,b=1,hex="|cff007fff"},
		purple = {r=1,g=0.4,b=1,hex="|cffff66ff"},
	}
	
	self.opts = {
		type = "group",
		args = {
			Columns = {
				name = L["Columns"],
				type = "range",
				desc = L["Number of Columns shown in the bag frames"],
				order = 1,
				get = function() return self.db.profile.columns end,
				set = function(value) self.db.profile.columns = value self:UpdateBags() end,
				min = 2,
				max = 20,
				step = 1,
			},
			Layout = {
				name = L["Layout"],
				type = "group",
				desc = L["Layout of the bag frames."],
				order = 2,
				args = {
					[L["Automatic"]] = {
						name = L["Automatic"],
						type = "toggle",
						desc = L["Automatically arrange the bag frames as the default ui does"],
						get = function() return self.db.profile.layout == "auto" end,
						set = function() self.db.profile.layout = "auto" end,
					},
					[L["Manual"]] = {
						name = L["Manual"],
						type = "toggle",
						desc = L["Each bag frame can be positioned manually."],
						get = function() return self.db.profile.layout == "manual" end,
						set = function() self.db.profile.layout = "manual" end,
					},
				},
			},
			Scale = {
				name = L["Scale"],
				type = "range",
				desc = L["Scale of the bag frames"],
				order = 2,
				max = 2,
				min = 0.3,
				step = 0.1,
				get = function() return self.db.profile.scale end,
				set = function(value) self.db.profile.scale = value 	
							self:UpdateBagScale()
							if self.db.profile.layout == "auto" then
								self:LayoutBagFrames()
							end
						end,
			},				
			SetLayoutBounds = {
				name = L["Set Layout Bounds"],
				type = "execute",
				order = 3,
				desc = L["Shows a frame you can drag and size to set where the bags will be placed when Layout is automatic"],
				func = function() self:ShowPlacementFrame() end,
			},	
			Lock = {
				name = L["Lock"],
				type = "toggle",
				desc = L["Locks the bag frames making them unmovable"],
				order = 3,
				get = function() return self.db.profile.lock end,
				set = function(value) self.db.profile.lock = value end
			},	
			SectionTitle = {
				name = L["Show Section Title"],
				type = "toggle",
				desc = L["Show a title on each section of the bags"],
				order = 4,
				get = function() return self.db.profile.showsectiontitle end,
				set = function(value) self.db.profile.showsectiontitle = value 	
							self:UpdateBags()
						end
			},		
			HideEmptySections = {
				name = L["Hide Empty Sections"],
				type = "toggle",
				desc = L["Hide sections that have no items in them."],
				order = 5,
				get = function() return self.db.profile.hideemptysections end,
				set = function(value) self.db.profile.hideemptysections = value 	
							self:UpdateBags()
						end
			},	
			HideDefaultBank = {
				name = L["Hide Default Bank"],
				type = "toggle",
				desc = L["Hide the default bank window."],
				order = 5,
				get = function() return self.db.profile.hidedefaultbank end,
				set = function(value) self.db.profile.hidedefaultbank = value end,
			},
			ShrinkWidth = {
				name = L["Shrink Width"],
				type = "toggle",
				desc = L["Shrink the bag's width to fit the items contained in them"],
				order = 6,
				get = function() return self.db.profile.shrinkwidth or self.db.profile.optimizesectionlayout end,
				set = function(value) self.db.profile.shrinkwidth = value 	
							self:UpdateBags()
						end,
				disabled = function() return self.db.profile.optimizesectionlayout	end
			},	
			SectionLayout = {
				name = L["Optimize Section Layout"],
				type = "toggle",
				desc = L["Change order and layout of sections in order to save display space."],
				order = 7,
				get = function() return self.db.profile.optimizesectionlayout end,
				set = function(value) self.db.profile.optimizesectionlayout = value 	
							self:UpdateBags()
						end
			},
			Compress = {
				name = L["Compress"],
				desc = L["Compress Multiple stacks into one item button"],
				type = "group",
				order = 8,
				args = {
					CompressAll = {
						name = L["Compress All"],
						type = "toggle",
						desc = L["Show all items as a single button with a count on it"],
						order = 1,
						get = function() return self.db.profile.compressall end,
						set = function(value) self.db.profile.compressall = value 	
									self:UpdateBags()
								end,
					},	
					CompressEmptySlots = {
						name = L["Compress Empty Slots"],
						type = "toggle",
						desc = L["Show all empty slots as a single button with a count on it"],
						order = 2,
						disabled = function() return self.db.profile.compressall end,
						get = function() return self.db.profile.compressempty end,
						set = function(value) self.db.profile.compressempty = value 	
									self:UpdateBags()
								end,
					},	
					CompressShards = {
						name = L["Compress Soul Shards"],
						type = "toggle",
						desc = L["Show all soul shards as a single button with a count on it"],
						order = 3,
						disabled = function() return self.db.profile.compressall end,
						get = function() return self.db.profile.compressshards end,
						set = function(value) self.db.profile.compressshards = value 	
									self:UpdateBags()
								end,
					},	
					CompressAmmo = {
						name = L["Compress Ammo"],
						type = "toggle",
						desc = L["Show all ammo as a single button with a count on it"],
						order = 4,
						disabled = function() return self.db.profile.compressall end,
						get = function() return self.db.profile.compressammo end,
						set = function(value) self.db.profile.compressammo = value 	
									self:UpdateBags()
								end,
					},	
				}
			},
			QualityColor = {
				name = L["Quality Colors"],
				desc = L["Color item buttons based on the quality of the item"],
				type = "group",
				order = 9,
				args = {
					Enable = {
						name = L["Enable"],
						type = "toggle",
						desc = L["Enable quality coloring"],
						order = 1,
						get = function() return self.db.profile.qualitycolor end,
						set = function(value) self.db.profile.qualitycolor = value 	
									self:UpdateItemButtons()
								end,
					},	
					Threshold = {
						name = L["Color Threshold"],
						type = "range",
						desc = L["Only color items of this quality or above"],
						order = 2,
						max = 6,
						min = 0,
						step = 1,
						get = function() return self.db.profile.qualitycolormin end,
						set = function(value) self.db.profile.qualitycolormin = value 	
									self:UpdateItemButtons()
								end,
					},	
					Intensity = {
						name = L["Color Intensity"],
						type = "range",
						desc = L["Intensity of the quality coloring"],
						order = 3,
						max = 1,
						min = 0.1,
						step = 0.1,
						get = function() return self.db.profile.qualitycolorintensity end,
						set = function(value) self.db.profile.qualitycolorintensity = value 	
									self:UpdateItemButtons()
								end,
					},	
				}
			},
			Sort = {
				name = L["Sort"],
				type = "group",
				desc = L["How items are sorted"],
				order = 10,
				args = {
					[L["Quality"]] = {
						name = L["Quality"],
						type = "toggle",
						desc = L["Items are sorted by quality."],
						get = function() return self.db.profile.sort == "quality" end,
						set = function() self.db.profile.sort = "quality" end,
					},
					[L["Name"]] = {
						name = L["Name"],
						type = "toggle",
						desc = L["Items are sorted by name."],
						get = function() return self.db.profile.sort == "name" end,
						set = function() self.db.profile.sort = "name" end,
					},
				},
			},	
			FubarText = {
				name = L["FuBar Text"],
				type = "group",
				desc = L["Options for the text shown on fubar"],
				order = 11,
				args = {	
					ShowEmpty = {
						name = L["Show empty bag slots"],
						type = "toggle",
						order = 1,
						desc = L["Show empty bag slots"],
						get = function() return self.db.profile.showempty end,
						set = function(value) self.db.profile.showempty = value self:UpdateText() end,
					},
					ShowUsed = {
						name = L["Show used bag slots"],
						type = "toggle",
						order = 2,
						desc = L["Show used bag slots"],
						get = function() return self.db.profile.showused end,
						set = function(value) self.db.profile.showused = value self:UpdateText() end,
					},
					ShowTotal = {
						name = L["Show Total bag slots"],
						type = "toggle",
						order = 3,
						desc = L["Show Total bag slots"],
						get = function() return self.db.profile.showtotal end,
						set = function(value) self.db.profile.showtotal = value self:UpdateText() end,
					},
					Combine = {
						name = L["Combine Counts"],
						type = "toggle",
						order = 10,
						desc = L["Show only one count with all the seclected types included"],
						get = function() return self.db.profile.combinecounts end,
						set = function(value) self.db.profile.combinecounts = value self:UpdateText() end,
					},
					ShowAmmo = {
						name = L["Show Ammo Bags Count"],
						type = "toggle",
						order = 10,
						desc = L["Show Ammo Bags Count"],
						get = function() return self.db.profile.showammocount end,
						set = function(value) self.db.profile.showammocount = value self:UpdateText() end,
					},
					ShowSoul = {
						name = L["Show Soul Bags Count"],
						type = "toggle",
						order = 10,
						desc = L["Show Soul Bags Count"],
						get = function() return self.db.profile.showsoulcount end,
						set = function(value) self.db.profile.showsoulcount = value self:UpdateText() end,
					},
					ShowSpecialty = {
						name = L["Show Specialty Bags Count"],
						type = "toggle",
						order = 3,
						desc = L["Show Specialty (profession etc) Bags Count"],
						get = function() return self.db.profile.showspecialcount end,
						set = function(value) self.db.profile.showspecialcount = value self:UpdateText() end,
					},
				},
			},
			HideDuplicates = {
				name = L["Hide Duplicate Items"],
				type = "toggle",
				desc = L["Prevents items from appearing in more than one section/bag."],
				order = 12,
				get = function() return self.db.profile.hideduplicates end,
				set = function(value) self.db.profile.hideduplicates = value 	
							self:UpdateBags()
						end
			},
			HighlightNew = {
				name = L["Highlight New Items"],
				type = "toggle",
				desc = L["Add *New* to new items, *+++* to items that you have gained more of."],
				order = 13,
				get = function() return self.db.profile.highlightnew end,
				set = function(value) self.db.profile.highlightnew = value 	
							self:UpdateItemButtons()
						end
			},
			ResetNew = {
				name = L["Reset New Items"],
				type = "execute",
				desc = L["Resets the new items highlights."],
				order = 14,
				func = function() self:SaveItemCounts() self:UpdateItemButtons() end,
			},
			Skin = {
				name = "Bag Skin", -- TODO: LOCALE
				type = "text",
				desc = "Select bag skin", -- TODO: LOCALE
				order = 15,
				get = function() return self.db.profile.skin end,
				set = 'ApplySkin',
				validate = self:GetSkinList()
			},
			spacer1 = {
				type = "header",
				order = 49,
			},				
			EditBags = {
				name = L["Edit Bags"],
				type = "execute",
				order = 50,
				desc = L["Edit the Bag Definitions"],
				func = function() self:OpenBagEditTablet() end,
			},
			EditCategories = {
				name = L["Edit Categories"],
				type = "execute",
				order = 51,
				desc = L["Edit the Category Definitions"],
				func = function() self:OpenCategoryEditTablet() end,
			},
			spacer2 = {
				type = "header",
				order = 99,
			},			
			LoadProfile = {
				name = L["Load Profile"],
				type = "group",
				desc = L["Load a built-in profile: NOTE: ALL Custom Bags will be lost and any edited built in categories will be lost."],
				order = 100,
				args = {
					Default = {
						name = L["Default"],
						type = "execute",
						desc = L["A default set of bags sorting your inventory into categories"],
						func = function() self:ApplyProfile(self.profiles.default)	end,
					},
					AllInOne = {
						name = L["All in one"],
						type = "execute",
						desc = L["A single bag containing your whole inventory, sorted by quality"],
						func = function() self:ApplyProfile(self.profiles.allinone)	end,
					},
					AllInOneSorted = {
						name = L["All In One Sorted"],
						type = "execute",
						desc = L["A single bag containing your whole inventory, sorted into categories"],
						func = function() self:ApplyProfile(self.profiles.allinonesorted) end,
					},
				},
			},				
	
		}			
	}
	
	self.OnMenuRequest = self.opts
	
	self:RegisterDB("BagginsDB")
	self:RegisterDefaults('profile', {
		showsectiontitle = true,
		hideemptysections = true,
		hidedefaultbank = true,
		compressempty = true,
		compressshards = true,
		compressammo = true,
		compressall = false,
		shrinkwidth = true,
		showspecialcount = true,
		showammocount = true,
		showsoulcount = true,
		showempty = true,
		showused = true,
		showtotal = true,
		combinecounts = false,
		highlightnew = true,
		hideduplicates = false,
		optimizesectionlayout = false,
		skin = 'default',
		scale = 1,
		rightoffset = 50,
		topoffset = 50,
		bottomoffset = 50,
		qualitycolorintensity = 0.3,
		qualitycolor = true,
		qualitycolormin = 2,
		columns = 5,
		sort = "quality",
		layout = "auto",
		bags = { },
		categories = {},
	}
	)
	
	self.profiles = {
		allinone = {
			layout = "manual",
			columns = 12,
			sorttype = "quality",
			showsectiontitle = true,
			bags = { 
				{ 
					name = L["All In One"],
					openWithAll = true,
					sections = 
					{ 
						{ name=L["Bags"], cats={L["Bags"]} }, 
						{ name=L["Ammo Bag"], cats={L["AmmoBag"]} }, 
						{ name=L["Soul Bag"], cats={L["SoulBag"]} }, 
						{ name=L["KeyRing"], cats={L["KeyRing"]} } 
					} 
				},
				{ 
					name = L["Bank All In One"],
					openWithAll = true,
					isBank = true,
					sections = 
					{ 
						{ name=L["Bank Bags"], cats={L["BankBags"]} },
					} 
				}
			}
		},
		allinonesorted = {
			layout = "manual",
			columns = 12,
			sorttype = "quality",
			showsectiontitle = true,
			optimizesectionlayout = true,
			bags = { 
				{ 
					name = L["All In One"],
					openWithAll = true,
					sections = 
					{
						{ name = L["Ammo"], cats = { L["AmmoBag"] } },
						{ name = L["SoulBag"], cats = { L["SoulBag"] } },
						{ name = L["KeyRing"], cats = { L["KeyRing"] },},
						{ name = L["Armor"], cats = { L["Armor"] },},
						{ name = L["Weapons"], cats = { L["Weapons"] } },
						{ name = L["Consumables"], cats = { L["Consumables"] } },
						{ name = L["Quest"], cats = { L["Quest"] } },
						{ name = L["Trade Goods"], cats = { L["Tradeskill Mats"], L["Gathered"] } },
						{ name = L["Other"], cats = { L["Other"] } }, 
					} 
				},
				{ 
					name = L["Bank All In One"],
					openWithAll = true,
					isBank = true,
					sections = 
					{ 
						{ name = L["Bank Equipment"], cats = { L["Armor"], L["Weapons"] },},
						{ name = L["Bank Consumables"], cats = { L["Consumables"] } },
						{ name = L["Bank Quest"], cats = { L["Quest"] } },
						{ name = L["Bank Trade Goods"], cats = { L["Tradeskill Mats"], L["Gathered"] } },
						{ name = L["Bank Other"], cats = { L["Other"] } }, 
					} 
				}
			}
		},
		default = {
			showsectiontitle = true,
			columns = 5,
			sorttype = "quality",
			layout = "auto",
			bags = {
				{
					name = L["Equipment"],
					openWithAll = true,
					sections = 
					{ 
						{ name=L["Armor"], cats={ L["Armor"] } }, 
						{ name=L["Weapons"], cats={ L["Weapons"] } } 
					}
				},
				{
					name = L["Quest"],
					openWithAll = true,
					sections = 
					{ 
						{ name=L["Quest Items"], cats = { L["Quest"] } } 
					}
				},	
				{
					name = L["Consumables"],
					openWithAll = true,
					sections = 
					{
						{ name = L["Water"], cats = {L["Water"]}}, 
						{ name = L["Food"], cats = {L["Food"]}},
						{ name = L["First Aid"], cats = {L["FirstAid"]}},
						{ name = L["Potions"], cats = {L["Potions"]}},
						{ name = L["Scrolls"], cats = {L["Scrolls"]}},
						{ name = L["Misc"], cats = { L["Misc Consumables"] }},
					}
				},
				{
					name = L["Trade Goods"],
					openWithAll = true,
					sections = 
					{
						{ name=L["Mats"], cats={ L["Tradeskill Mats"] } },
						{ name=L["Gathered"], cats={ L["Gathered"] } },
					}
				},
				{
					name = L["Ammo"],
					openWithAll = true,
					sections = 
					{ 
						{ name=L["Ammo"], cats={ L["AmmoBag"] } }, 
						{ name=L["SoulShards"], cats={ L["SoulBag"] } } 
					}
				},		
				{
					name = L["Other"],
					openWithAll = true,
					sections = 
					{ 
						{ name=L["Other"], cats = { L["Other"] } }, 
						{ name=L["Trash"], cats = { L["Trash"], L["TrashEquip"] } }, 
						{ name=L["Empty"], cats = { L["Empty"] } } 
					}
				},
				{
					name = L["Bank Equipment"],
					openWithAll = true,
					isBank = true,
					sections = 
					{ 
						{ name=L["Armor"], cats={ L["Armor"] } }, 
						{ name=L["Weapons"], cats={ L["Weapons"] } } 
					}
				},
				{
					name = L["Bank Quest"],
					openWithAll = true,
					isBank = true,
					sections = 
					{ 
						{ name=L["Quest Items"], cats = { L["Quest"] } } 
					}
				},	
				{
					name = L["Bank Consumables"],
					openWithAll = true,
					isBank = true,
					sections = 
					{
						{ name = L["Water"], cats = {L["Water"]}}, 
						{ name = L["Food"], cats = {L["Food"]}},
						{ name = L["First Aid"], cats = {L["FirstAid"]}},
						{ name = L["Potions"], cats = {L["Potions"]}},
						{ name = L["Scrolls"], cats = {L["Scrolls"]}},
						{ name = L["Misc"], cats = { L["Misc Consumables"] }},
					}
				},
				{
					name = L["Bank Trade Goods"],
					openWithAll = true,
					isBank = true,
					sections = 
					{
						{ name=L["Mats"], cats={ L["Tradeskill Mats"] } },
						{ name=L["Gathered"], cats={ L["Gathered"] } },
					}
				},
				{
					name = L["Bank Other"],
					openWithAll = true,
					isBank = true,
					sections = 
					{ 
						{ name=L["Other"], cats = { L["Other"] } }, 
						{ name=L["Trash"], cats = { L["Trash"], L["TrashEquip"] } }, 
						{ name=L["Empty"], cats = { L["Empty"] } } 
					}
				},				
			}
		}
	}
	
	self.defaultcategories = {
		[L["Misc Consumables"]] = { name=L["Misc Consumables"], {type="ItemType" ,itype = "Consumable"},{operation = "NOT",type = "PTSet",setname = "All Edible Food",},{operation = "NOT",type = "PTSet",setname = "Water",},{operation = "NOT",type = "PTSet",setname = "Potion",},{operation = "NOT",type = "PTSet",setname = "Scrolls",},{operation = "NOT",type = "PTSet",setname = "Bandages",},},
		[L["Consumables"]] = { name=L["Consumables"], {type="ItemType" ,itype = "Consumable"} },
		[L["Armor"]] = { name=L["Armor"], { type="ItemType", itype="Armor" } },
		[L["Weapons"]] = { name=L["Weapons"], { type="ItemType", itype="Weapon" }, { type="PTSet", setname="Tradeskill Tools", operation="NOT" }, },
		[L["Quest"]] = { name=L["Quest"], { type="ItemType", itype="Quest" }, { type="Tooltip", text="ITEM_BIND_QUEST" } },
		[L["AmmoBag"]] = { name=L["AmmoBag"], { type="AmmoBag" } },
		[L["Trash"]] = { name=L["Trash"], { type="Quality", quality = 0, comp = "<=" } },
		[L["TrashEquip"]] = { name=L["TrashEquip"], { type="ItemType", itype="Armor" }, { type="ItemType", itype="Weapon", operation="OR" }, { type="Quality", quality = 1, comp = "<=", operation="AND" }, { type="PTSet", setname="Tradeskill Tools", operation="NOT" }, { type="ItemType", itype="Quest", operation="NOT" }, },
		[L["Other"]] = { name=L["Other"], { type="Other" } },
		[L["Empty"]] = { name=L["Empty"], { type="Empty" }, { type="AmmoBag", operation="NOT"}, { type="ContainerType", ctype="Soul Bag", operation="NOT"}, },
		[L["Bags"]] = { name=L["Bags"], { type="Bag", bagid=1 }, { type="Bag", bagid=2, operation="OR" }, { type="Bag", bagid=3, operation="OR" }, { type="Bag", bagid=4, operation="OR" }, { type="Bag", bagid=0, operation="OR" }, { type="AmmoBag", operation="NOT"}, { type="ContainerType", ctype="Soul Bag", operation="NOT"} },
		[L["BankBags"]] = { name=L["BankBags"], { type="Bag", bagid=-1 }, { type="Bag", bagid=5, operation="OR" }, { type="Bag", bagid=6, operation="OR" }, { type="Bag", bagid=7, operation="OR" }, { type="Bag", bagid=8, operation="OR" }, { type="Bag", bagid=9, operation="OR" }, { type="Bag", bagid=10, operation="OR" }, { type="Bag", bagid=11, operation="OR" }, },
		[L["KeyRing"]] = { name=L["KeyRing"], { type="Bag", noempty=true, bagid=-2 }, },
		[L["Potions"]] = { name=L["Potions"], { type="PTSet", setname="Potion" }, },
		[L["Food"]] = { name=L["Food"], { type="PTSet", setname="All Edible Food" }, },
		[L["Scrolls"]] = { name=L["Scrolls"], { type="PTSet", setname="Scrolls" }, },
		[L["FirstAid"]] = { name=L["FirstAid"], { type="PTSet", setname="CraftedBy - First Aid" }, },
		[L["Water"]] = { name=L["Water"], { type="PTSet", setname="Water" }, },
		[L["SoulBag"]] = { name=L["SoulBag"], { type="ContainerType", ctype="Soul Bag" }, },
		[L["Tradeskill Mats"]] = { name=L["Tradeskill Mats"], { type="PTSet", setname="Tradeskill Mats" }, },
		[L["Gathered"]] = { name=L["Gathered"], { type="PTSet", setname="Gather" }, },
	}

	self:RegisterEditTablets()
	
	self:BuildPTSetTable()
	
	self:RegisterChatCommand({ "/baggins" }, self.opts )
end

function Baggins:OnEnable(firstload)
	self:RegisterBucketEvent("BAG_UPDATE",0.1, "OnBagUpdate")
	self:RegisterEvent("BAG_UPDATE_COOLDOWN", "UpdateItemButtonCooldowns")
	self:RegisterEvent("ITEM_LOCK_CHANGED", "UpdateItemButtonLocks")
	self:RegisterEvent("PLAYERBANKSLOTS_CHANGED", "OnBankChanged")
	self:RegisterEvent("BANKFRAME_CLOSED", "OnBankClosed")
	self:RegisterEvent("BANKFRAME_OPENED", "OnBankOpened")
	self:RegisterEvent('AUCTION_HOUSE_SHOW', "OpenAllBags")
	self:RegisterEvent('AUCTION_HOUSE_CLOSED', "CloseAllBags")
	self:RegisterEvent('Baggins_RefreshBags');
	
	self:Hook("OpenBackpack", true)
	self:Hook("CloseBackpack", true)
	self:Hook("CloseSpecialWindows", true)
	self:Hook("BankFrame_OnEvent", true)
	--force an update of all bags on first opening
	self.doInitialUpdate = true
	self.doInitialBankUpdate = true
	
	--self:SetDebugging(true);
	
	self:EnableSkin(self.db.profile.skin)
	self:OnProfileEnable()
end

function Baggins:OnProfileDisable()
	self:CloseAllBags()
end

function Baggins:OnProfileEnable()
	local p = self.db.profile
	--check if this profile has been setup before, if not add the default bags and categories
	--cant leave these in the defaults since removing a bag would have it come back on reload
	local catsexist
	for k in pairs(p.categories) do
		catsexist = true
		break
	end
	if not catsexist then
		deepCopy(p.categories, self.defaultcategories)
	end
	 if #p.bags == 0 then
		deepCopy(p.bags, self.profiles.default.bags)
	end	
	
	for k, v in ipairs(p.bags) do
		if not self.bagframes[k] then
			self:CreateBagFrame(k)
		end
	end
	
	--convert old ItemID rules to tables
	for catname, cat in pairs(self.db.profile.categories) do
		for ruleid, rule in pairs(cat) do
			if rule.ids and type(rule.ids) == "string" then
				local tmp = {(" "):split(rule.ids)}
				rule.ids = {}
				for i, v in pairs(tmp) do
					rule.ids[tonumber(v)] = true
				end
			end
		end
	end
end

function Baggins:BankFrame_OnEvent(event)
	if not self:IsActive() or not self.db.profile.hidedefaultbank then
		self.hooks.BankFrame_OnEvent(event)
	end
end

function Baggins:OnDisable()
	self:CloseAllBags()
end

function Baggins:SaveItemCounts()
	for k in pairs(itemcounts) do
		itemcounts[k] = nil
	end
	for bag = 0, 4 do
		for slot = 1, GetContainerNumSlots(bag) do
			local link = GetContainerItemLink(bag, slot)
			if link then
				local id = tonumber(link:match("item:(%d+)"))
				if id then
					itemcounts[id] = GetItemCount(id)
				end
			end
		end
	end
end

function Baggins:IsCompressed(itemID) 
	local p = self.db.profile
	if p.compressall then 
		return true 
	end
	--string id's here are empty slots
	if type(itemID) == "string" and p.compressempty then
		return true 
	end
	if type(itemID) == "number" then
		if p.compressshards and itemID == 6265 then
			return true
		end
		if p.compressammo and pt:ItemInSet(itemID,"Ammo") then
			return true
		end
	end
end

function Baggins:IsSpecialBag(bag)
	if not bag then return end
	if type(bag) == "string" then bag = tonumber(bag) end
	if BagTypes[bag] == 2 then
		return "b"
	end
	if BagTypes[bag] == 3 then
		return "k"
	end

	local link = GetInventoryItemLink("player",19+bag)
	if link then
		local _, _, _, _, _, Type, SubType = GetItemInfo(link)
		if Type == L["Quiver"] then
			return "a"
		end
		if Type == L["Container"] then
			if SubType == L["Bag"] then
				return
			end
			if SubType == L["Soul Bag"] then 
				return "s"
			end
			if SubType == L["Enchanting Bag"] then
				return "n"
			end
			if SubType == L["Engineering Bag"] then
				return "e"
			end
			if SubType == L["Herb Bag"] then
				return "h"
			end
		end
	end
	
end

function Baggins:OnBankClosed()
	-- don't remove the test, it prevents infinite recursion loop on CloseBankFrame()
	if self.bankIsOpen then 
		self.bankIsOpen = nil
		self:CloseAllBags()
	end
end

function Baggins:OnBankOpened()
	if self.doInitialBankUpdate then
		self.doInitialBankUpdate = nil
		self:ForceFullBankUpdate()
	end
	self.bankIsOpen = true
	self:OpenAllBags()
end

function Baggins:OnBankChanged()
	self:OnBagUpdate(-1)
end

function Baggins:ApplyProfile(profile)
	local p = self.db.profile
	self:CloseAllBags()
	for k, v in pairs(profile) do  
		if type(v) == "table" then
			if not p[k] then
				p[k] = {}
			end
			deepCopy(p[k],v)
		else
			p[k] = v
		end
	end 
	for k, v in pairs(self.defaultcategories) do
		if not p.categories[k] then
			p.categories[k] = {}
		end
		deepCopy(p.categories[k], v)
	end
	tablet:Refresh("BagginsEditCategories")
	tablet:Refresh("BagginsEditBags")
	self:ForceFullUpdate()
	self:OpenAllBags()
end

-------------------------
-- Update Bag Contents --
-------------------------
local scheduled_refresh = false;

function Baggins:ScheduleRefresh()
	if not scheduled_refresh then
		self:Debug('Scheduling refresh');
		scheduled_refresh = self:ScheduleEvent('Baggins_RefreshBags', 0);
	end
end

function Baggins:Baggins_RefreshBags()
	
	if self.dirtyBags then
		self:Debug('Updating bags');
		self:ReallyUpdateBags();
	end
	for bagid,bagframe in pairs(self.bagframes) do
		
		for secid,sectionframe in pairs(bagframe.sections) do
			if sectionframe.used and sectionframe.dirty then
				self:Debug('Updating section #%d-%d', bagid, secid);
				self:ReallyLayoutSection(sectionframe)
			end
		end
		
		if bagframe.dirty then
			self:ReallyUpdateBagFrameSize(bagid)
		end
	end
	if self.dirtyBagLayout then
		self:Debug('Updating bag layout');
		self:ReallyLayoutBagFrames()
	end

	self:Debug('Refresh done');
	scheduled_refresh = nil;
end

function Baggins:UpdateBags()
	self.dirtyBags = true;
	self:ScheduleRefresh();
end

local othersections = {}

function Baggins:ReallyUpdateBags()
	local p = self.db.profile
	self:StartBags()
	local isVisible = false
	for bagid, bag in pairs(p.bags) do
		if self.bagframes[bagid] and self.bagframes[bagid]:IsVisible() then
			isVisible = true
		end
	end
	if not isVisible then return end
	for bagid, bag in pairs(p.bags) do	
		if ( not bag.isBank ) or self.bankIsOpen then
			if self.bagframes[bagid] then
				local cache, color
				if bag.isBank then
					cache = bankcategorycache
					color = self.colors.blue
				else
					cache = categorycache
					color = self.colors.black
				end
				self:StartBag(self.bagframes[bagid])
				self:SetBagTitle(bagid,bag.name)
				self.currentSkin:SetBankVisual(self.bagframes[bagid], bag.isBank);
				
				for sectionid, section in pairs(bag.sections) do
					self:StartSection()
					for catid, catname in pairs(section.cats) do
						if not cache[catname] then
							cache[catname] = {}
						end
						if not cache[catname]["Other"] then
							self:AppendCategory(catname,bag.isBank)
						else
							othersections[bagid] = sectionid
						end
					end
					if self.bagframes[bagid]:IsVisible() then						
						self:SetSection(bagid,sectionid,section.name,self:FinishSection())
					end
				end
				self:UpdateBagFrameSize(bagid)
			end
		end
	end
	for bagid, sectionid in pairs(othersections) do
		self:StartSection()
		self:AppendOther(p.bags[bagid].isBank)
		self:SetSection(bagid,sectionid,p.bags[bagid].sections[sectionid].name,self:FinishSection())
		self:UpdateBagFrameSize(bagid)
		othersections[bagid] = nil
	end
	self.dirtyBags = nil;	
end

function Baggins:SetSection(bagid, secid,title, contents)
	local bagframe = self.bagframes[bagid]
	if not bagframe then return end
	local i
	for i = 1,secid do
		if not bagframe.sections[i] then
			bagframe.sections[i] = self:CreateSectionFrame(bagframe,i)
			if i == 1 then
				bagframe.sections[i]:SetPoint("TOPLEFT",bagframe,"TOPLEFT",10,-36)
			else
				bagframe.sections[i]:SetPoint("TOPLEFT",bagframe.sections[i-1],"BOTTOMLEFT",0,1)
			end
		end
	end
	self:UpdateSectionContents(bagframe.sections[secid],title,contents)
	self:UpdateBagFrameSize(bagid)
end

-----------------
-- Bag Updates --
-----------------
local firstbagupdate = true
function Baggins:OnBagUpdate(bagids)
	if firstbagupdate then
		firstbagupdate = false
		self:SaveItemCounts()
	end
	self:UpdateText()
	local itemschanged
	
	if type(bagids) == "number" then
		itemschanged = self:CheckSlotsChanged(bagids)
	else
		for bag in pairs(bagids) do
			itemschanged = self:CheckSlotsChanged(bag) or itemschanged
		end
	end
	if itemschanged then
		self:UpdateBags()
	else
		self:UpdateItemButtons()
	end
end

function Baggins:ForceFullUpdate()
	--local start = GetTime()
	local bagid
	for bagid = 0, 4 do
		self:CheckSlotsChanged(bagid, true)
	end
	self:CheckSlotsChanged(-2,true)
	self:UpdateBags()
	--self:Print("Full Update took "..(math.floor((GetTime()-start)*1000)/1000).." seconds")
end

function Baggins:ForceFullBankUpdate()
	local bagid
	for bagid = 5, 11 do
		self:CheckSlotsChanged(bagid, true)
	end
	self:CheckSlotsChanged(-1,true)
end

function Baggins:CheckSlotsChanged(bag, forceupdate)
	local slot
	local itemschanged
	for slot = 1, GetContainerNumSlots(bag) do
		local key = bag..":"..slot
		local iteminfo
		
		local link = GetContainerItemLink(bag, slot)
		local _, itemCount = GetContainerItemInfo(bag, slot)
		local itemid, itemName
		if link then
			itemid = link:match("item:(%d+)")
			itemName = GetItemInfo(link)
			iteminfo = itemid.." "..itemCount.." "..(itemName or "_")
		end
		
		if slotcache[key] ~= iteminfo or forceupdate then
			local olditemid = (slotcache[key] or ""):match("^(%d+)")
			if itemid ~= olditemid then
				itemschanged = true
			end
			self:OnSlotChanged(bag, slot)
			slotcache[key] = iteminfo
		end
	end
	return itemschanged
end

function Baggins:OnSlotChanged(bag, slot)
	local isbank
	local cache
	if BagTypes[bag] == 2 then
		used = bankuseditems
		cache = bankcategorycache
		isbank = true
	else
		used = useditems
		cache = categorycache
	end
	
	local key = bag..":"..slot
	if not used[key] then
		used[key] = false
	end
	for catid, category in pairs(self.db.profile.categories) do
		if self:CategoryInUse(catid) then
			if not cache[catid] then
				cache[catid] = {}
			end
			cache[catid][key] = nil

			local catmatch = false
			cache[catid]["Other"] = nil
			for ruleid, rule in ipairs(category) do
				local rulematch
				if rule.type == "Other" then
					cache[catid]["Other"] = true
				else
					if ruleid == 1 then 
						operation = "OR"
					else
						operation = rule.operation or "OR"
					end
					rulematch = (RuleTypes[rule.type] and RuleTypes[rule.type].Matches(bag,slot,rule))
					if operation == "OR" then
						catmatch = catmatch or rulematch
					elseif operation == "AND" then
						catmatch = catmatch and rulematch
					elseif operation == "NOT" then
						catmatch = catmatch and (not rulematch)
					end
				end
			end
			if catmatch then
				cache[catid][key] = true
			end
		end
	end
end

-----------------------------
-- Update Section Contents --
-----------------------------
function Baggins:UpdateSectionContents(sectionframe,title,itemlist)
	local p = self.db.profile
	local cols = p.columns
	local usedStack = false
	local stackType = "*"
	
	self:SortItemList(itemlist,p.sort)
	local itemcount = 0
	
	for k, v in ipairs(sectionframe.items) do
		v.used = nil
	end
	for k, item in ipairs(itemlist) do
		if itemlist[item] then
			local bag, slot = string.match(item,"^(-?%d+):(%d+)$")
			local bagType = self:IsSpecialBag(bag)
			local itemID
			if slotcache[item] then
				itemID = tonumber(string.match(slotcache[item],"^(%d+)"))
			else
				itemID = self:IsSpecialBag(bag) or ""
			end
			local isStack = false
			if self:IsCompressed(itemID) then
				isStack = true
				if stackType ~= itemID then
					stackType = itemID
					usedStack = false
				end
			end
			if bag and slot then
				bag = tonumber(bag)
				slot = tonumber(slot)
				
				if not isStack or (isStack and not usedStack) then
					sectionframe.items[itemcount+1] = sectionframe.items[itemcount+1] or self.ItemButtonPool:Acquire()
					local itemframe = sectionframe.items[itemcount+1]
					if isStack then 
						usedStack = true
						itemframe.countoverride = true
					else
						itemframe.countoverride = nil
					end
					self:UpdateItemButton(sectionframe:GetParent(),itemframe,bag,slot)
					
					itemframe.used = true
					itemcount = itemcount + 1
				end
			end
		end
	end
	sectionframe.itemcount = itemcount
	sectionframe.used = true
	sectionframe:Show()

	self:LayoutSection(sectionframe, title)
end
local function NameComp(a, b)
	local baga, slota = string.match(a,"^(-?%d+):(%d+)$")
	local bagb, slotb = string.match(b,"^(-?%d+):(%d+)$")
	local linka = GetContainerItemLink(baga, slota)
	local linkb = GetContainerItemLink(bagb, slotb)
	--if both are empty slots compare based on bag type
	if not (linka or linkb) then
		bagtypea = Baggins:IsSpecialBag(baga)
		bagtypeb = Baggins:IsSpecialBag(bagb)
		if not bagtypea then
			return false
		end
		if not bagtypeb then
			return true
		end
		return bagtypea < bagtypeb
	end
	if not linka then
		return false
	end
	if not linkb then
		return true
	end
	local namea = GetItemInfo(linka)
	local nameb = GetItemInfo(linkb)
	if not namea then return true end
	if not nameb then return false end		
	if namea == nameb then
		_, counta = GetContainerItemInfo(baga, slota)
		_, countb = GetContainerItemInfo(bagb, slotb)
		return counta < countb
	else
		return namea < nameb
	end
end
local function QualityComp(a, b)
	local baga, slota = string.match(a,"^(-?%d+):(%d+)$")
	local bagb, slotb = string.match(b,"^(-?%d+):(%d+)$")
	local linka = GetContainerItemLink(baga, slota)
	local linkb = GetContainerItemLink(bagb, slotb)
	--if both are empty slots compare based on bag type
	if not (linka or linkb) then
		bagtypea = Baggins:IsSpecialBag(baga)
		bagtypeb = Baggins:IsSpecialBag(bagb)
		if not bagtypea then
			return false
		end
		if not bagtypeb then
			return true
		end
		return bagtypea < bagtypeb
	end
	if not linka then
		return false
	end
	if not linkb then
		return true
	end

	local namea, _, quala = GetItemInfo(linka)
	local nameb, _, qualb = GetItemInfo(linkb)
	if not quala then return true end
	if not qualb then return false end
	if quala == qualb then
		if namea == nameb then
			_, counta = GetContainerItemInfo(baga, slota)
			_, countb = GetContainerItemInfo(bagb, slotb)
			return counta > countb
		else
			return namea < nameb
		end
	else
		return quala < qualb
	end
end
function Baggins:SortItemList(itemlist, sorttype)
	local func
	if sorttype == "name" then
		func = NameComp
	elseif sorttype == "quality" then
		func = QualityComp 
	else
		return
	end
	table.sort(itemlist,func)
end


------------------------
-- Section/Bag Layout --
------------------------
function Baggins:GetSectionSize(sectionframe, maxcols)
	local count = sectionframe.itemcount
	maxcols = maxcols or count
	local width = math.min(count, maxcols) * 39
	local height = maxcols > 0 and (math.ceil(count / maxcols) * 39) or 0
	
	if self.db.profile.showsectiontitle then
		width = math.max(width, sectionframe.title:GetWidth())
		height = height + sectionframe.title:GetHeight()
	end
		
	return width, height
end

local areas = {}

function Baggins:OptimizeSectionLayout(bagid)
	local bagframe = self.bagframes[bagid]
	if not bagframe then
		return
	end
	local p = self.db.profile;
	local s = self.currentSkin;
	local titlefactor = p.showsectiontitle and 1 or 0;
	local totalwidth, totalheight = 0, 0
		
	for k in pairs(areas) do areas[k] = nil; end
	
	table.insert(areas, format('0:0:%d:1000', self.db.profile.columns * 39))
	
	--self:Debug("**** Laying out bag #%d ***", bagid)
	
	for secid,sectionframe in ipairs(bagframe.sections) do		
		local count = sectionframe.itemcount;
		if sectionframe.used and (count > 0 or not p.hideemptysections) then					
			local minwidth = self:GetSectionSize(sectionframe, 1);
			local _, minheight = self:GetSectionSize(sectionframe);

			--[[self:Debug("Section #%d, %d item(s), title width: %q, min width: %q, min height: %q",
				secid,	sectionframe.itemcount, sectionframe.title:GetWidth(), minwidth,
				minheight);]]
						
			sectionframe.layout_waste = nil;
			sectionframe.layout_columns = nil;
			sectionframe.layout_area_index = nil;
			
			for areaid,area in pairs(areas) do			
				--self:Debug("  Area #%d: %s", areaid, area)
				
				local area_w,area_h = area:match('^%d+:%d+:(%d+):(%d+)$');
				area_w = tonumber(area_w)
				area_h = tonumber(area_h)
				
				if area_w >= minwidth and area_h >= minheight then
					local cols = math.floor(area_w / 39)
					local width, height = self:GetSectionSize(sectionframe, cols);
					--self:Debug("    %d columns", cols);
					
					if area_h >= height then
						local waste = (area_w * area_h) - (width * height)
						--self:Debug("    area waste: %d", waste)
						
						if not sectionframe.layout_waste or waste < sectionframe.layout_waste then
							sectionframe.layout_waste = waste
							sectionframe.layout_columns = cols
							sectionframe.layout_areaid = areaid
							--self:Debug("    -> best fit")
						else
							--self:Debug("    -> do not fit")
						end												
						
					else
						--self:Debug("  -> too short")
					end					
					
				else
					--self:Debug("  -> too small")
				end
				
			end			
			
			local areaid = sectionframe.layout_areaid;
			local area_x, area_y, area_w, area_h = table.remove(areas, areaid):match('^(%d+):(%d+):(%d+):(%d+)$');
			area_x = tonumber(area_x)
			area_y = tonumber(area_y)
			area_w = tonumber(area_w)
			area_h = tonumber(area_h)
			local cols = sectionframe.layout_columns;
			local width, height = self:GetSectionSize(sectionframe, cols);
			sectionframe:SetPoint('TOPLEFT', bagframe, 'TOPLEFT', s.BagLeftPadding + area_x, - (s.BagTopPadding + area_y) );
			sectionframe:SetWidth(width)
			sectionframe:SetHeight(height)
			
			totalwidth = math.max(totalwidth, area_x + width)
			totalheight = math.max(totalheight,area_y + height)
			self:ReallyLayoutSection(sectionframe, cols);
			
			width = width + 19
			height = height + 10
			if area_w - width >= 39 then					
				local area = format('%d:%d:%d:%d', area_x + width, area_y, area_w - width, height);
				--self:Debug("Created new area: %s", area)
				table.insert(areas, area)
			end			
			if area_h - height >= 39 then
				local area = format('%d:%d:%d:%d', area_x, area_y + height, area_w, area_h - height);
				--self:Debug("Created new area: %s", area)
				table.insert(areas, area)
			end
		end		
	end	
	
	return s.BagLeftPadding + s.BagRightPadding + totalwidth, s.BagTopPadding + s.BagBottomPadding + totalheight
end

function Baggins:UpdateBagFrameSize(bagid)
	local bagframe = self.bagframes[bagid]
	if not bagframe then return end
	bagframe.dirty = true;
	self:ScheduleRefresh();	
end

function Baggins:ReallyUpdateBagFrameSize(bagid)
	local bagframe = self.bagframes[bagid]
	if not bagframe then return end
	local p = self.db.profile
	local s = self.currentSkin

	self:Debug('Updating bag #%d', bagid);
	
	local hpadding = s.BagLeftPadding + s.BagRightPadding	
	local width = bagframe.title:GetWidth() + s.BagLeftPadding + s.TitlePadding
	local height = s.BagTopPadding + s.BagBottomPadding

	if self.db.profile.optimizesectionlayout then
		local swidth, sheight = Baggins:OptimizeSectionLayout(bagid)		
		width = math.max(width, swidth)
		height = math.max(height, sheight)
	else
		for id,section in pairs(bagframe.sections) do
			if section.used then
				if id == 1 then
					section:SetPoint("TOPLEFT",bagframe,"TOPLEFT",s.BagLeftPadding,-s.BagTopPadding)
				else
					section:SetPoint("TOPLEFT",bagframe.sections[id-1],"BOTTOMLEFT",0,-10)
					height = height + 10
				end
				self:ReallyLayoutSection(section)
				width = math.max(width,section:GetWidth()+hpadding)
				height = height + section:GetHeight()
			end
		end
		if not p.shrinkwidth then
			width = math.max(p.columns*39 + hpadding, width)
		end
	end
	
	bagframe:SetWidth(width)
	bagframe:SetHeight(height)		
	
	bagframe.dirty = nil;	
end

function Baggins:LayoutSection(sectionframe, title, cols)
	sectionframe.dirty = true;
	sectionframe.set_title = title;
	sectionframe.set_columns = cols;
	self:ScheduleRefresh();
end

function Baggins:ReallyLayoutSection(sectionframe, cols)
	local p = self.db.profile
	local totalwidth, totalheight = 1,1
	local s = self.currentSkin
	cols = cols or sectionframe.set_columns or p.columns
	
	if p.showsectiontitle then
		sectionframe.title:SetText(sectionframe.set_title)
		totalheight = totalheight + s.SectionTitleHeight
	else
		sectionframe.title:SetText("")
	end
	
	for itemnum, itemframe in ipairs(sectionframe.items) do
		if itemframe.used then
			if itemnum == 1 then
				if p.showsectiontitle then
					itemframe:SetPoint("TOPLEFT",sectionframe.title,"BOTTOMLEFT",0,0)
				else
					itemframe:SetPoint("TOPLEFT",sectionframe,"TOPLEFT",0,0)
				end
				totalwidth = 39
				totalheight = totalheight+39
			elseif itemnum%cols == 1 then
				itemframe:SetPoint("TOPLEFT",sectionframe.items[itemnum-cols],"BOTTOMLEFT",0,-2)
				totalheight = totalheight+39
			else
				itemframe:SetPoint("TOPLEFT",sectionframe.items[itemnum-1],"TOPRIGHT",2,0)
				if itemnum <= cols then
					totalwidth = totalwidth+39
				end
			end
			itemframe:Show()
		else
			itemframe:Hide()
		end
	end
	
	if p.showtitle then
		totalwidth = max(totalwidth,sectionframe.title:GetWidth())
	end
	
	if sectionframe.itemcount == 0 and p.hideemptysections then
		totalheight = 1
		sectionframe:Hide()
	end
	sectionframe:SetWidth(totalwidth)
	sectionframe:SetHeight(totalheight)	
	
	sectionframe.dirty = nil;
end

----------------
-- Skin stuff --
----------------

do -- private
	local skins = {}
	local function void() end

	function Baggins:GetSkinList()
		local skinlist = {}
		for name in pairs(skins) do
			table.insert(skinlist, name)
		end
		return skinlist
	end
	
	function Baggins:GetSkin(name)
		return skins[name]
	end
	
	function Baggins:RegisterSkin(name, skin)
		skin.core = self
		skin.name = skin.name or name
		skin.SkinSection = skin.SkinSection or void
		skin.SkinItem = skin.SkinItem or void
		skin.SkinBag = skin.SkinBag or void
		skin.UnskinSection = skin.UnskinSection or void
		skin.UnskinItem = skin.UnskinItem or void
		skin.UnskinBag = skin.UnskinBag or void
		skins[name] = skin
	end
		
end

function Baggins:DisableSkin(name)
	local oldskin = self:GetSkin(name);
	if oldskin and self.currentSkin == oldskin then		
		for bagid,bagframe in ipairs(self.bagframes) do
			for secid,sectionframe in ipairs(bagframe.sections) do
				for itemid,itemframe in ipairs(sectionframe.items) do
					oldskin:UnskinItem(itemframe)
				end
				oldskin:UnskinSection(sectionframe)
			end
			oldskin:UnskinBag(bagframe);
		end
		self.currentSkin = nil
	end
end

function Baggins:EnableSkin(name)
	local newskin = self:GetSkin(name);
	if newskin and self.currentSkin ~= newskin then
		for bagid,bagframe in ipairs(self.bagframes) do
			for secid,sectionframe in ipairs(bagframe.sections) do
				for itemid,itemframe in ipairs(sectionframe.items) do
					newskin:SkinItem(itemframe)
				end
				newskin:SkinSection(sectionframe)
				sectionframe.dirty = true
			end
			newskin:SkinBag(bagframe);
			bagframe.dirty = true
		end
		self.currentSkin = newskin
		self:UpdateBags()
	end
end

function Baggins:ApplySkin(name)
	if name ~= self.db.profile.skin and self:GetSkin(name) then		
		self:DisableSkin(self.db.profile.skin)
		self.db.profile.skin = name		
		self:EnableSkin(name)
	end
end

------------------
-- Default Skin --
------------------
 
local defaultSkin = {

	BagLeftPadding = 10,
	BagRightPadding = 10,
	BagTopPadding = 32,
	BagBottomPadding = 10,
	TitlePadding = 32,
	SectionTitleHeight = 13,
	
	EmptySlotTexture = false,

	BagFrameBackdrop = {
		bgFile = "Interface/Tooltips/UI-Tooltip-Background", 
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border", 
		tile = true, tileSize = 16, edgeSize = 16, 
		insets = { left = 5, right = 5, top = 5, bottom = 5 }
	},
	
	NormalBagColor = 'black',
	BankBagColor = 'blue',
	
}
	
function defaultSkin:SkinBag(frame)
	frame:SetBackdrop(self.BagFrameBackdrop)

	frame.closebutton:SetPoint("TOPRIGHT",frame,"TOPRIGHT",0,0)
	
	frame.title:SetVertexColor(1,1,1,1)
	frame.title:SetPoint("TOPLEFT",frame,"TOPLEFT",10,-10)
	frame.title:SetHeight(12)	
end

function defaultSkin:SetBankVisual(frame, isBank)
	local color = self.core.colors[self.NormalBagColor]
	if isBank then
		color = self.core.colors[self.BankBagColor]
	end
	frame:SetBackdropColor(color.r, color.g, color.b)
end

function defaultSkin:SkinSection(frame)
	frame.title:SetVertexColor(1,1,1,1)
	frame.title:SetText("")
	frame.title:SetPoint("TOPLEFT",frame,"TOPLEFT",0,0)
	frame.title:SetHeight(13)
end

------------------------
-- Blizzard-like Skin --
------------------------

Baggins:RegisterSkin('default', defaultSkin);

local blizzardSkin = setmetatable({

	EmptySlotTexture = 'Interface\\AddOns\\Baggins\\Textures\\EmptySlot',

	BagLeftPadding = 12,
	BagRightPadding = 11,
	BagTopPadding = 50,
	BagBottomPadding = 11,
	TitlePadding = 70,

	BagFrameBackdrop = {
		bgFile = 'Interface\\AddOns\\Baggins\\Textures\\BagFrameBackground', 
		edgeFile = 'Interface\\AddOns\\Baggins\\Textures\\BagFrameBorder', 
		tile = true, tileSize = 32, edgeSize = 32, 
		insets = { left = 7, right = 6, top = 7, bottom = 6 }
	},

	NormalBagColor = 'white',

	JustifyH = {},

}, {__index = defaultSkin}) -- inherits from defaultSkin

function blizzardSkin:SkinBag(frame)
	frame:SetBackdrop(self.BagFrameBackdrop)
	
	frame.closebutton:SetPoint("TOPRIGHT",frame,"TOPRIGHT",0,-1)
	
	frame.title:SetVertexColor(1,1,1,1)
	frame.title:SetPoint("TOPRIGHT",frame,"TOPRIGHT",-33,-9)
	frame.title:SetHeight(12)	
	self.JustifyH[frame.title] = frame.title:GetJustifyH()
	frame.title:SetJustifyH('RIGHT')
	
	if not frame.portrait then
		frame.portrait = frame:CreateTexture(frame:GetName().."Portrait","OVERLAY")
		frame.portrait:SetWidth(48)
		frame.portrait:SetHeight(48)
		frame.portrait:SetPoint("TOPLEFT", frame, "TOPLEFT", -4, 0)
		frame.portrait:SetTexture('Interface\\AddOns\\Baggins\\Textures\\BagFramePortraitFrame')
		frame.portrait:SetTexCoord(0, 0.745, 0, 0.745)
	end
	frame.portrait:Show()
	
	if not frame.icon then
		frame.icon = frame:CreateTexture(frame:GetName().."Icon","ARTWORK")
		frame.icon:SetWidth(34)
		frame.icon:SetHeight(34)
		frame.icon:SetPoint("TOPLEFT", frame, "TOPLEFT", 3, -5)
		frame.icon:SetTexture("Interface\\Icons\\INV_Jewelry_Ring_03")
	end
	frame.icon:Show()
	
end

function blizzardSkin:UnskinBag(frame)
	frame.portrait:Hide()
	frame.icon:Hide()
	frame.title:SetJustifyH(self.JustifyH[frame.title])
	self.JustifyH[frame.title] = nil
end

function blizzardSkin:SetBankVisual(frame, isBank)
	local color = self.core.colors[self.NormalBagColor]
	if isBank then
		color = self.core.colors[self.BankBagColor]
	end
	frame:SetBackdropColor(color.r, color.g, color.b)
	frame:SetBackdropBorderColor(color.r, color.g, color.b)
	frame.portrait:SetVertexColor(color.r, color.g, color.b)
end


Baggins:RegisterSkin('blizzard', blizzardSkin);

--------------------
-- Item Filtering --
--------------------
function Baggins:OnRuleChanged()
	--self:ForceFullUpdate()
	self:UpdateBags()
	tablet:Refresh("BagginsEditCategories")
end

function Baggins:StartBags()
	for k, v in pairs(useditems) do
		useditems[k] = false
	end
	for k, v in pairs(bankuseditems) do	
		bankuseditems[k] = false
	end
end

--start filling a bag, sets all sections to hidden untill they are filled
function Baggins:StartBag(bag)
	for sectionid, sectionframe in ipairs(bag.sections) do
		sectionframe.used = false
		sectionframe:Hide()
		for itemid,itemframe in ipairs(sectionframe.items) do
			itemframe:Hide()
		end
	end
end

local itemstable = {}
--starting a new section, clear the itemstable
function Baggins:StartSection()
	self:EmptyItemsTable()
end

function Baggins:EmptyItemsTable()
	for k, v in pairs(itemstable) do
		itemstable[k] = nil
	end
end

--finished a section, return the item list
function Baggins:FinishSection()
	return itemstable
end

--appends a category to the itemslist
function Baggins:AppendCategory(catname,isbank)
	if isbank then 
		cache = bankcategorycache
		used = bankuseditems
	else
		cache = categorycache
		used = useditems
	end

	if not cache[catname] then return end
	if not cache[catname]["Other"] then
		for k, v in pairs(cache[catname]) do
			if itemstable[k] == nil and not ( used[k] and self.db.profile.hideduplicates ) then
				tinsert(itemstable, k)
				itemstable[k] = true
				used[k] = true
			end
		end
	end
end

function Baggins:AppendOther(isbank)
	local used
	if isbank then 
		used = bankuseditems
	else
		used = useditems
	end
	for k, v in pairs(used) do
		if not v then
			local bag = string.match(k, "(-?%d+):")
			if bag then
				bag = tonumber(bag)
				--if this isnt the keyring
				if BagTypes[bag] ~= 3 then
					tinsert(itemstable, k)
					itemstable[k] = true
				end
			end
		end
	end
end

function Baggins:GetIncludeRule(category,create)
	local numrules = #category
	for i = numrules,1,-1 do
		if category[i].type == "ItemID" then
			if category[i].operation ~= "NOT" then
				return category[i]
			end
		end		
	end
	if create then
		local newrule = { type="ItemID", operation="OR"}
		tinsert( category, newrule )
		return newrule
	end
end

function Baggins:GetExcludeRule(category,create)
	local numrules = #category
	for i = numrules,1,-1 do
		if category[i].type == "ItemID" then
			if category[i].operation == "NOT" then
				return category[i]
			end
		end		
	end
	if create then
		local newrule = { type="ItemID", operation="NOT"}
		tinsert( category, newrule )
		return newrule
	end
end

function Baggins:IncludeItemInCategory(catname, itemid)
	local cat = self.db.profile.categories[catname]
	if not cat then return end
	
	for i, rule in ipairs(cat) do
		if rule.ids and rule.operation == "NOT" then
			rule.ids[itemid] = nil
		end
	end
	local rule = self:GetIncludeRule(cat,true)
	if not rule.ids then
		rule.ids = {}
	end
	rule.ids[itemid] = true
	tablet:Refresh("BagginsEditCategories") 
	if not self.CatEditOpen then
		self:ForceFullUpdate()
	end
end

function Baggins:ExcludeItemFromCategory(catname, itemid)
	local cat = self.db.profile.categories[catname]
	if not cat then return end
	
	for i, rule in ipairs(cat) do
		if rule.ids and rule.operation ~= "NOT" then
			rule.ids[itemid] = nil
		end
	end
	local rule = self:GetExcludeRule(cat,true)
	if not rule.ids then
		rule.ids = {}
	end
	rule.ids[itemid] = true
	tablet:Refresh("BagginsEditCategories") 
	if not self.CatEditOpen then
		self:ForceFullUpdate()
	end	
end

-----------------
-- Slot Counts --
-----------------
function Baggins:CountEmptySlots(bagtype)
	local bag, slot
	local count = 0
	if bagtype == "b" then
		return self:CountEmptyBankSlots()
	end
	if bagtype == "k" then
		for slot = 1, GetContainerNumSlots(-2) do
			if not GetContainerItemLink(-2,slot) then
				count = count + 1
			end
		end
		return count
	end
	
	for bag = 0, 4 do
		if bagtype == self:IsSpecialBag(bag) or bagtype == "ALL" then
			for slot = 1, GetContainerNumSlots(bag) do
				if not GetContainerItemLink(bag,slot) then
					count = count + 1
				end
			end
		end
	end
	return count
end

function Baggins:CountSlots(bagtype)
	local bag, slot
	local count = 0
	if bagtype == "b" then
		return self:CountBankSlots()
	end
	if bagtype == "k" then
		for slot = 1, GetContainerNumSlots(-2) do
			count = count + 1
		end
		return count
	end
	
	for bag = 0, 4 do
		if bagtype == self:IsSpecialBag(bag) or bagtype == "ALL" then
			for slot = 1, GetContainerNumSlots(bag) do
				count = count + 1
			end
		end
	end
	return count
end

function Baggins:CountEmptyBankSlots()
	local bag, slot
	local count = 0
	for bag = -1, 11 do
		if BagTypes[bag] == 2 then
			for slot = 1, GetContainerNumSlots(bag) do
				if not GetContainerItemLink(bag,slot) then
					count = count + 1
				end
			end
		end
	end
	return count
end

function Baggins:CountBankSlots()
	local bag, slot
	local count = 0
	for bag = -1, 11 do
		if BagTypes[bag] == 2 then
			for slot = 1, GetContainerNumSlots(bag) do
				count = count + 1
			end
		end
	end
	return count
end


---------------------
-- Frame Creation  --
---------------------
function Baggins:CreateBagPlacementFrame()
local f = CreateFrame("frame","BagginsBagPlacement",UIParent)

	f:SetWidth(130)
	f:SetHeight(300)
	f:SetPoint("CENTER",UIParent,"CENTER",0,0)
	f:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background", 
	                                            edgeFile = false, 
	                                            edgeSize = 0, 
	                                            insets = { left = 0, right = 0, top = 0, bottom = 0 }})
	f:SetBackdropColor(0.2,0.5,1,0.5)
		
	f:EnableMouse(true)
	f:SetMovable(true)
	f:SetResizable(true)
	f:SetClampedToScreen(true)
	f:SetScript("OnMouseDown",function(this, button) if button == "RightButton" then this:Hide() else this:StartMoving() end end)
	f:SetScript("OnMouseUp", function() this:StopMovingOrSizing() self:SaveBagPlacement() end)

	f.t = CreateFrame("frame","BagginsBagPlacementTopMover",f)
	f.t:SetPoint("TOPLEFT",f,"TOPLEFT",0,0)
	f.t:SetPoint("TOPRIGHT",f,"TOPRIGHT",0,0)
	f.t:SetHeight(20)
	f.t:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background", 
	                                            edgeFile = false, 
	                                            edgeSize = 0, 
	                                            insets = { left = 0, right = 0, top = 0, bottom = 0 }})
	f.t:SetBackdropColor(0,0,1,1)
	f.t:EnableMouse(true)
	f.t:SetScript("OnMouseDown",function() this:GetParent():StartSizing("TOP") end)
	f.t:SetScript("OnMouseUp", function() this:GetParent():StopMovingOrSizing() self:SaveBagPlacement() end)
	
	f.b = CreateFrame("frame","BagginsBagPlacementTopMover",f)
	f.b:SetPoint("BOTTOMLEFT",f,"BOTTOMLEFT",0,0)
	f.b:SetPoint("BOTTOMRIGHT",f,"BOTTOMRIGHT",0,0)
	f.b:SetHeight(20)
	f.b:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background", 
	                                            edgeFile = false, 
	                                            edgeSize = 0, 
	                                            insets = { left = 0, right = 0, top = 0, bottom = 0 }})
	f.b:SetBackdropColor(0,0,1,1)
	f.b:EnableMouse(true)
	f.b:SetScript("OnMouseDown",function() this:GetParent():StartSizing("BOTTOM") end)
	f.b:SetScript("OnMouseUp", function() this:GetParent():StopMovingOrSizing() self:SaveBagPlacement() end)
	
	f.midtext = f:CreateFontString(nil,"OVERLAY","GameFontNormalSmall")
	f.midtext:SetText(L["Drag to Move\nRight-Click to Close"])
	f.midtext:SetPoint("LEFT",f,"LEFT",0,0)
	f.midtext:SetPoint("RIGHT",f,"RIGHT",0,0)
	f.midtext:SetHeight(45)
	f.midtext:SetVertexColor(1, 1, 1)
	
	f.toptext = f.t:CreateFontString(nil,"OVERLAY","GameFontNormalSmall")
	f.toptext:SetText(L["Drag to Size"])
	f.toptext:SetPoint("TOPLEFT",f,"TOPLEFT",0,-2)
	f.toptext:SetPoint("TOPRIGHT",f,"TOPRIGHT",0,-2)
	f.toptext:SetHeight(15)
	f.toptext:SetVertexColor(1, 1, 1)

	f.bottext = f.b:CreateFontString(nil,"OVERLAY","GameFontNormalSmall")
	f.bottext:SetText(L["Drag to Size"])
	f.bottext:SetPoint("BOTTOMLEFT",f,"BOTTOMLEFT",0,2)
	f.bottext:SetPoint("BOTTOMRIGHT",f,"BOTTOMRIGHT",0,2)
	f.bottext:SetHeight(15)
	f.bottext:SetVertexColor(1, 1, 1)	
end

local function BagginsItemButton_OnEnter(button)
	if ( not button ) then
		button = this
	end

	local x
	x = button:GetRight()
	if ( x >= ( GetScreenWidth() / 2 ) ) then
		GameTooltip:SetOwner(button, "ANCHOR_LEFT")
	else
		GameTooltip:SetOwner(button, "ANCHOR_RIGHT")
	end

	-- Keyring specific code
	if ( this:GetParent():GetID() == KEYRING_CONTAINER ) then
		GameTooltip:SetInventoryItem("player", KeyRingButtonIDToInvSlotID(this:GetID()))
		CursorUpdate()
		return
	end
	
	if this:GetParent():GetID() == -1 then
		
		GameTooltip:SetInventoryItem("player", BankButtonIDToInvSlotID(this:GetID()))
		CursorUpdate()
		return		
	end

	local hasCooldown, repairCost = GameTooltip:SetBagItem(button:GetParent():GetID(),button:GetID())
	
	if ( hasCooldown ) then
		button.updateTooltip = TOOLTIP_UPDATE_TIME
	else
		button.updateTooltip = nil
	end

	if ( InRepairMode() and (repairCost and repairCost > 0) ) then
		GameTooltip:AddLine(TEXT(REPAIR_COST), "", 1, 1, 1)
		SetTooltipMoney(GameTooltip, repairCost)
		GameTooltip:Show()
	elseif ( MerchantFrame:IsShown() and MerchantFrame.selectedTab == 1 ) then
		ShowContainerSellCursor(button:GetParent():GetID(),button:GetID())
	elseif ( this.readable or (IsControlKeyDown() and button.hasItem) ) then
		ShowInspectCursor()
	else
		ResetCursor()
	end
		
end

local function BagginsItemButton_PreClick(button)
	if IsAltKeyDown() and GetMouseButtonClicked() == "RightButton" then
		local link = GetContainerItemLink(button:GetParent():GetID(),button:GetID())
		if link then
			local itemid = link:match("item:(%d+)")
			if itemid then
				itemid = tonumber(itemid)
				dewdrop:Open( button,
					"children",function(level, value, valueN_1) 
						if level == 1 then
							dewdrop:AddLine("text",L["Add To Category"], "hasArrow", true, "value", "Add")
							dewdrop:AddLine("text",L["Exclude From Category"], "hasArrow", true, "value", "Exclude")
							dewdrop:AddLine("text",L["Item Info"], "hasArrow", true, "value", "Info")
						elseif level == 2 then
							local categories = Baggins.db.profile.categories
							while #catsorttable > 0 do
								tremove(catsorttable,#catsorttable)
							end
							for catid in pairs(categories) do
								tinsert(catsorttable,catid)
							end
							table.sort(catsorttable)
							
							if value == "Add" then
								for i, name in ipairs(catsorttable) do
									dewdrop:AddLine("text",name, "func", function(cat, item) Baggins:IncludeItemInCategory(cat, item) end, "arg1", name,"arg2", itemid)
								end
							elseif value == "Exclude" then
								for i, name in ipairs(catsorttable) do
									dewdrop:AddLine("text",name, "func", function(cat, item) Baggins:ExcludeItemFromCategory(cat, item) end, "arg1", name,"arg2", itemid)
								end
							elseif value == "Info" then
								local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemCount, itemEquipLoc, itemTexture = GetItemInfo(itemid)
								dewdrop:AddLine("text",L["ItemID: "]..itemid)
								dewdrop:AddLine("text",L["Item Type: "]..itemType)
								dewdrop:AddLine("text",L["Item Subtype: "]..itemSubType)								
								dewdrop:AddLine("text",L["Quality: "]..itemRarity)								
								dewdrop:AddLine("text",L["Level: "]..itemLevel)															
								dewdrop:AddLine("text",L["MinLevel: "]..itemMinLevel)									
								dewdrop:AddLine("text",L["Stack Size: "]..itemCount)									
								dewdrop:AddLine("text",L["Equip Location: "]..itemEquipLoc)		
								dewdrop:AddLine("text",L["Periodic Table Sets"], "hasArrow", true, "value", "PTSets")
							end
						elseif level == 3 then
							if value == "PTSets" then
								for k,v in pairs(pt.k) do
									for a,b in pairs(v) do
										if type(b) == "string" then
											if string.match(" "..b.." ", " "..itemid.."[%s:]") then
												dewdrop:AddLine('text',a)
											end
										end
									end
								end
							end
						end
					end)
			end
		end			
	end
end

----------------------
-- Frame pool class --
----------------------

local FramePool = AceLibrary('AceOO-2.0').Class()

local function Frame_ReleaseToPool(frame)
	if frame.__pool then
		frame.__pool:Release(frame)
	end
end

function FramePool.prototype:init(create, setup, cleanup)
	FramePool.super.prototype.init(self)
	self.pool = {}
	self.count = 0
	self.create = create
	self.setup = setup
	self.cleanup = cleanup
end

function FramePool.prototype:Acquire(...)
	local frame = next(self.pool)
	if frame then
		self.pool[frame] = nil
		--Baggins:Debug('reusing old frame %q', frame)
	else
		self.count = self.count + 1
		frame = self.create(self.count)
		--Baggins:Debug('creating new frame %q', frame)
	end
	frame:ClearAllPoints()
	frame.__pool = self
	frame.ReleaseToPool = Frame_ReleaseToPool	
	if self.setup then
		self.setup(frame, ...)
	end
	return frame
end

function FramePool.prototype:Release(frame)
	if not frame then
		return
	end
	if frame.__pool ~= self then
		AceLibrary:error('trying to release a frame to the wrong frame pool')
	end
	--Baggins:Debug('releasing frame %q', frame)
	frame:Hide()
	frame:ClearAllPoints()
	frame:UnregisterAllEvents()
	if self.cleanup then
		self.cleanup(frame)
	end
	self.pool[frame] = nil
end

----------------------
-- Item button pool --
----------------------

Baggins.ItemButtonPool = FramePool:new(
	-- create
	function(number)
		local frame = CreateFrame("Button","BagginsItem"..number,nil,"ContainerFrameItemButtonTemplate")
		
		frame.glow = frame:CreateTexture(nil,"OVERLAY")
		frame.glow:SetTexture("Interface\\Addons\\Baggins\\Textures\\Glow")
		frame.glow:SetAlpha(0.3)
		frame.glow:SetAllPoints(frame)
		
		frame.newtext = frame:CreateFontString(frame:GetName().."NewText","OVERLAY","GameFontNormal")
		frame.newtext:SetPoint("TOP",frame,"TOP",0,0)
		frame.newtext:SetHeight(13)
		frame.newtext:SetTextColor(0,1,0,1)
		frame.newtext:SetShadowColor(0,0,0,1)
		frame.newtext:SetShadowOffset(1,-1)
		frame.newtext:SetText("*New*")
		frame.newtext:Hide()
	
		local cooldown = getglobal(frame:GetName().."Cooldown")
		cooldown:SetAllPoints(frame)
		cooldown:SetFrameLevel(10)
		
		frame:SetFrameStrata("HIGH")
		frame:SetScript("OnEnter",BagginsItemButton_OnEnter)
		frame:SetScript("PreClick",BagginsItemButton_PreClick)
		
		return frame
	end,
	-- setup
	function(frame)
		Baggins.currentSkin:SkinItem(frame)
	end,
	-- cleanup
	function(frame)
		Baggins.currentSkin:UnskinItem(frame)
	end
)

function Baggins:CreateSectionFrame(bagframe,secid)
	local frame = CreateFrame("Frame",bagframe:GetName().."Section"..secid,bagframe)
	
	frame.title = frame:CreateFontString(bagframe:GetName().."SectionTitle","OVERLAY","GameFontNormalSmall")
	frame:SetFrameStrata("HIGH")
	frame:Show()
	frame.items = {}
	self.currentSkin:SkinSection(frame)
	
	return frame
end

function Baggins:CreateBagFrame(bagid)
	if not bagid then return end
	local bagname = "BagginsBag"..bagid
	
	local frame = CreateFrame("Frame",bagname,UIParent)
	self.bagframes[bagid] = frame
	frame.bagid = bagid
	frame:SetWidth(100)
	frame:SetHeight(100)
	frame:SetPoint("CENTER",UIParent,"CENTER",0,0)
	frame:EnableMouse(true)
	frame:SetMovable(true)
	frame:SetScript("OnMouseDown",function() if not self.db.profile.lock then this:StartMoving() end end)
	frame:SetScript("OnMouseUp",function() this:StopMovingOrSizing() self:SaveBagPosition(this.bagid) end)
	
	frame.closebutton = CreateFrame("Button",bagname.."CloseButton",frame,"UIPanelCloseButton")
	frame.closebutton:SetScript("OnClick", function()
		if IsShiftKeyDown() then
			self:CloseAllBags()
		else
			self:CloseBag(this:GetParent().bagid) 
		end
	end)
	
	frame.title = frame:CreateFontString(bagname.."Title","OVERLAY","GameFontNormalLarge")
	frame.title:SetText("Baggins")
	frame:SetFrameStrata("HIGH")
	frame:SetClampedToScreen(true) 
	frame:SetScale(self.db.profile.scale)
	frame.sections = {}
	frame.bagparents = {}
	
	self.currentSkin:SkinBag(frame)

	self:UpdateBagFrameSize(bagid)
	frame:Hide()
end

function Baggins:SetBagTitle(bagid,title)
	if self.bagframes[bagid] then
		self.bagframes[bagid].title:SetText(title)
		self:UpdateBagFrameSize(bagid)
	end
end


----------------------
-- ItemButton Funcs --
----------------------
function Baggins:UpdateItemButtons()
	for bagid, bag in ipairs(self.bagframes) do
		for sectionid, section in ipairs(bag.sections) do
			for buttonid, button in ipairs(section.items) do
				if button:IsVisible() then
					self:UpdateItemButton(bag,button,button:GetParent():GetID(),button:GetID())
				end
			end
		end
	end
end

function Baggins:UpdateItemButtonLocks()
	for bagid, bag in ipairs(self.bagframes) do
		for sectionid, section in ipairs(bag.sections) do
			for buttonid, button in ipairs(section.items) do
				if button:IsVisible() then
					local _, _, locked = GetContainerItemInfo(button:GetParent():GetID(), button:GetID())
					SetItemButtonDesaturated(button, locked, 0.5, 0.5, 0.5)
				end
			end
		end
	end
end

function Baggins:UpdateItemButtonCooldowns()
	for bagid, bag in ipairs(self.bagframes) do
		for sectionid, section in ipairs(bag.sections) do
			for buttonid, button in ipairs(section.items) do
				if button:IsVisible() then
					local container = button:GetParent():GetID()
					local slot = button:GetID()
					local texture = GetContainerItemInfo(container, slot)
					if ( texture ) then
						self:UpdateItemButtonCooldown(container, button)
						button.hasItem = 1
					else
						getglobal(button:GetName().."Cooldown"):Hide()
						button.hasItem = nil
					end
				end
			end
		end
	end
end

function Baggins:SetItemButtonCount(button, count, realcount)
	if not button then
		return
	end
	if not count then
		count = 0
	end
	local counttext = getglobal(button:GetName().."Count")
	if button.countoverride then
		button.count = realcount
	else
		button.count = count
	end
	if type(count) == "string" then
		counttext:ClearAllPoints()
		counttext:SetPoint("BOTTOMRIGHT",button,"BOTTOMRIGHT",-5,2)	
		counttext:SetText(count)
		counttext:Show()
	elseif ( count > 1 or (button.isBag and count > 0) ) then
		if ( count > 999 ) then
			count = (math.floor(count/100)/10).."k"
			counttext:ClearAllPoints()
			counttext:SetPoint("BOTTOM",button,"BOTTOM",0,2)
		else
			counttext:ClearAllPoints()
			counttext:SetPoint("BOTTOMRIGHT",button,"BOTTOMRIGHT",-5,2)		
		end
		counttext:SetText(count)
		counttext:Show()
	else
		counttext:Hide()
	end
end

function Baggins:UpdateItemButton(bagframe,button,bag,slot)
	local p = self.db.profile
	local texture, itemCount, locked, quality, readable = GetContainerItemInfo(bag, slot)
	local link = GetContainerItemLink(bag, slot)
	local itemid
	if link then
		local _,_,Qual = GetItemInfo(link)
		quality = Qual or quality
		itemid = tonumber(link:match("item:(%d+)"))
	end
	button:SetID(slot)
	if p.qualitycolor and texture and quality >= p.qualitycolormin then
		local r, g, b = GetItemQualityColor(quality)
		button.glow:SetVertexColor(r,g,b)
		button.glow:SetAlpha(p.qualitycolorintensity)
		button.glow:Show()
	else
		button.glow:Hide()
	end
	local text = button.newtext
	if p.highlightnew and itemid and bag >= 0 and bag <= 4 then
		local savedcount = itemcounts[itemid] 
		if not savedcount then
			text:SetText(L["*New*"])
			text:Show()
		else
			local count = GetItemCount(itemid)
			if count > savedcount then
				text:SetText("*+++*")
				text:Show()
			else
				text:Hide()
			end
		end
	else
		text:Hide()
	end
	
	if not bagframe.bagparents[bag] then 
		bagframe.bagparents[bag] = CreateFrame("Frame",nil,bagframe)
		bagframe.bagparents[bag]:SetID(bag)
	end
	button:SetParent(bagframe.bagparents[bag])
	if texture then
		SetItemButtonTexture(button, texture)
	elseif self.currentSkin.EmptySlotTexture then
		SetItemButtonTexture(button, self.currentSkin.EmptySlotTexture)
	else
		SetItemButtonTexture(button, nil)
	end
	if button.countoverride then
		local count
		if not itemid then
			local bagtype = self:IsSpecialBag(bag)
			if bagtype then
				count = bagtype.."-"..self:CountEmptySlots(bagtype)
			else
				count = self:CountEmptySlots()
			end
		else
			if BagTypes[bag] == 2 then
				count = GetItemCount(itemid,true) - GetItemCount(itemid)
				if IsEquippedItem(itemid) then
					count = itemframe.countoverride - 1
				end
			else
				count = GetItemCount(itemid)
				if IsEquippedItem(itemid) then
					count = itemframe.countoverride - 1
				end
			end
		end
		self:SetItemButtonCount(button, count, itemCount)
	else
		self:SetItemButtonCount(button, itemCount)
	end
	SetItemButtonDesaturated(button, locked, 0.5, 0.5, 0.5)
	
	if ( texture ) then
		self:UpdateItemButtonCooldown(bag, button)
		button.hasItem = 1
	else
		getglobal(button:GetName().."Cooldown"):Hide()
		button.hasItem = nil
	end

	button.readable = readable
	--local normalTexture = getglobal(name.."Item"..j.."NormalTexture")
	--if ( quality and quality ~= -1) then
	--	local color = getglobal("ITEM_QUALITY".. quality .."_COLOR")
	--	normalTexture:SetVertexColor(color.r, color.g, color.b)
	--else
	--	normalTexture:SetVertexColor(TOOLTIP_DEFAULT_COLOR.r, TOOLTIP_DEFAULT_COLOR.g, TOOLTIP_DEFAULT_COLOR.b)
	--end
	local showSell = nil
	if ( GameTooltip:IsOwned(button) ) then
		if ( texture ) then
			local hasCooldown, repairCost = GameTooltip:SetBagItem(bag,slot)
			if ( hasCooldown ) then
				button.updateTooltip = TOOLTIP_UPDATE_TIME
			else
				button.updateTooltip = nil
			end
			
			if ( InRepairMode() and (repairCost > 0) ) then
				GameTooltip:AddLine(TEXT(REPAIR_COST), "", 1, 1, 1)
				SetTooltipMoney(GameTooltip, repairCost)
				GameTooltip:Show()
			elseif ( MerchantFrame:IsShown() and not locked) then
				showSell = 1
			end
		else
			GameTooltip:Hide()
		end
		if ( showSell ) then
			ShowContainerSellCursor(bag, slot)
		elseif ( readable ) then
			ShowInspectCursor()
		else
			ResetCursor()
		end
	end
end

function Baggins:UpdateItemButtonCooldown(container, button)
	local cooldown = getglobal(button:GetName().."Cooldown")
	local start, duration, enable = GetContainerItemCooldown(container, button:GetID())
	CooldownFrame_SetTimer(cooldown, start, duration, enable)
	if ( duration > 0 and enable == 0 ) then
		SetItemButtonTextureVertexColor(button, 0.4, 0.4, 0.4)
	end
end


---------------------
-- Frame Placement --
---------------------
-- manual
function Baggins:SaveBagPosition(bagid)
	local frame = self.bagframes[bagid]
	if not frame then return end
	local s = frame:GetEffectiveScale()
	self.db.profile.bags[bagid].x = frame:GetLeft() * s
	self.db.profile.bags[bagid].y = frame:GetTop() * s
end

function Baggins:LoadBagPosition(bagid)	

	local bag = self.db.profile.bags[bagid]
	local frame = self.bagframes[bagid]
	if not frame then return end
	local s = frame:GetEffectiveScale()
	frame:ClearAllPoints()
	if bag.x and bag.y then
		frame:SetPoint("TOPLEFT",UIParent,"BOTTOMLEFT", bag.x / s, bag.y / s)
	else
		frame:SetPoint("CENTER",UIParent,"CENTER",0,0)
	end
end
-- auto
function Baggins:SaveBagPlacement()
	local p = self.db.profile
	local frame = BagginsBagPlacement
	if not frame then return end
	local s = frame:GetEffectiveScale()
	p.rightoffset = (GetScreenWidth() - frame:GetRight() )
	p.bottomoffset = frame:GetBottom()
	p.topoffset = (GetScreenHeight() - frame:GetTop() )
end

function Baggins:LoadBagPlacement()		
	local p = self.db.profile
	local frame = BagginsBagPlacement
	if not frame then return end
	local s = frame:GetEffectiveScale()
	frame:ClearAllPoints()
	frame:SetPoint("BOTTOMRIGHT",UIParent,"BOTTOMRIGHT", - p.rightoffset, p.bottomoffset)
	frame:SetPoint("TOPRIGHT",UIParent,"TOPRIGHT", - p.rightoffset, - p.topoffset)
end

function Baggins:ShowPlacementFrame()
	if not BagginsBagPlacement then
		self:CreateBagPlacementFrame()
	end
	Baggins:LoadBagPlacement()
	BagginsBagPlacement:Show()
end

function Baggins:LayoutBagFrames()
	self.dirtyBagLayout = true;
	self:ScheduleRefresh();
end

function Baggins:ReallyLayoutBagFrames()
	local p = self.db.profile
	local frame, xOffset, yOffset, screenHeight, freeScreenHeight, column, maxwidth
	local screenWidth = GetScreenWidth()

	screenHeight = GetScreenHeight()
	
	-- Adjust the start anchor for bags depending on the multibars
	xOffset = p.rightoffset
	yOffset = p.bottomoffset
	-- freeScreenHeight determines when to start a new column of bags
	freeScreenHeight = screenHeight - yOffset - p.topoffset
	maxwidth = 1
	for index, frame in ipairs(self.bagframes) do
		maxwidth = math.max(maxwidth, frame:GetWidth())
	end
	column = 0
	local index = 1
	local prevframe
	for id, frame in ipairs(self.bagframes) do
		if frame:IsVisible() then
			frame:ClearAllPoints()
			local s = frame:GetEffectiveScale() / UIParent:GetEffectiveScale()
			if ( index == 1 ) then
				-- First bag
				frame:SetPoint("BOTTOMRIGHT", frame:GetParent(), "BOTTOMRIGHT", -xOffset/s, yOffset/s )
				index = index + 1
				prevframe = frame
			elseif ( freeScreenHeight < frame:GetHeight()*s ) then
				-- Start a new column
				column = column + 1
				freeScreenHeight = screenHeight - yOffset - p.topoffset
				frame:SetPoint("BOTTOMRIGHT", frame:GetParent(), "BOTTOMRIGHT", -(column * maxwidth) - xOffset/s, yOffset/s )
				index = index + 1
				prevframe = frame
			else
				-- Anchor to the previous bag
				frame:SetPoint("BOTTOMRIGHT", prevframe, "TOPRIGHT", 0, CONTAINER_SPACING)	
				index = index + 1
				prevframe = frame
			end
			self:SaveBagPosition(id)
			freeScreenHeight = freeScreenHeight - frame:GetHeight()*s - VISIBLE_CONTAINER_SPACING
		end
	end
	self.dirtyBagLayout = nil
end

function Baggins:UpdateBagScale()
	for i, frame in ipairs(self.bagframes) do
		frame:SetScale(self.db.profile.scale)
	end
end	


---------------------
-- Edit Tablets    --
---------------------
function Baggins:RegisterEditTablets()
	tablet:Register("BagginsEditBags" ,
					"children", function() self:FillBagEditTablet() end,
					"data", {},
					"cantAttach", true,
					"clickable", true,
					"point", function() return "TOPLEFT", "TOPLEFT" end,
					"parent", "UIParent",
					"showTitleWhenDetached", true,
					"showHintWhenDetached", true,
					"dontHook", true,
					"strata","DIALOG",
					"menu", function() dewdrop:Close() end
					)				
	tablet:Register("BagginsEditCategories" ,
					"children", function() self:FillCategoryEditTablet() end,
					"data", {},
					"cantAttach", true,
					"clickable", true,
					"point", function() return "TOPLEFT", "TOPLEFT" end,
					"parent", "UIParent",
					"showTitleWhenDetached", true,
					"showHintWhenDetached", true,
					"dontHook", true,
					"strata","DIALOG",
					"menu", function() dewdrop:Close() end
					)		
	self.dewdropparent = CreateFrame("frame",nil,UIParent)
	self.dewdropparent:SetHeight(1)
	self.dewdropparent:SetWidth(1)
	self.dewdropparent:SetPoint("CENTER",UIParent,"CENTER",0,0)
end

function Baggins:OpenBagEditTablet()
	tablet:Open("BagginsEditBags")
end

function Baggins:OpenCategoryEditTablet()
	self.CatEditOpen = true
	tablet:Open("BagginsEditCategories")
end

function Baggins:FillBagEditTablet()
	local cat
	tablet:SetTitle(L["Bags"])
	tablet:SetTitleColor(1,1,1)
	for bagid, bag in ipairs(self.db.profile.bags) do
		local color
		if bag.isBank then
			color = self.colors.blue
		else
			color = self.colors.white
		end
		if currentbag ~= bagid then
			cat = tablet:AddCategory("id",bagid,"text","+ "..bag.name,"size",14,"showWithoutChildren",true,"hideBlankLine",true,
										"func",function(bagid) 
											if IsShiftKeyDown() then
												self:MoveBag(bagid)
											elseif IsAltKeyDown() then
												self:MoveBag(bagid, true)
											elseif IsControlKeyDown() then 
												self:StartDeleteBag(bagid) 
											else 
												currentbag = bagid
												currentsection = nil
											end 
											tablet:Refresh("BagginsEditBags") 
										end,
										"arg1",bagid,
										"textR",color.r,"textG",color.g,"textB",color.b)
		else
			cat = tablet:AddCategory("id",bagid,"text","- "..bag.name,"size",14,"showWithoutChildren",true,"hideBlankLine",true,
										"func",function(bagid) 
											if IsShiftKeyDown() then
												self:MoveBag(bagid)
											elseif IsAltKeyDown() then
												self:MoveBag(bagid, true)
											elseif IsControlKeyDown() then 
												self:StartDeleteBag(bagid) 
											else 
												currentbag = nil 
												currentsection = nil
											end 
											tablet:Refresh("BagginsEditBags") 
										end,
										"arg1",bagid,
										"textR",color.r,"textG",color.g,"textB",color.b)
										
			cat:AddLine("text",L["Options"],"size",12,"indentation",10)
			cat:AddLine("text",L["Open With All"],"size",11,"indentation",15,"func",function(bag) bag.openWithAll = not bag.openWithAll tablet:Refresh("BagginsEditBags") end,"arg1",bag,"hasCheck",true,"checked",bag.openWithAll,"textR",1,"textG",1,"textB",1)
			cat:AddLine("text",L["Bank"],"size",11,"indentation",15,"func",function(bag) bag.isBank = not bag.isBank tablet:Refresh("BagginsEditBags") self:UpdateBags() end,"arg1",bag,"hasCheck",true,"checked",bag.isBank,"textR",1,"textG",1,"textB",1)
			cat:AddLine("text",L["Sections"],"size",12,"indentation",10)
			for sectionid, section in ipairs(bag.sections) do
				if sectionid ~= currentsection then
					cat:AddLine("text","+ "..section.name,"size",12,"indentation", 20,"textR",1,"textG",1,"textB",1,
									"func",function(bagid,sectionid) 
										if IsShiftKeyDown() then
											self:MoveSection(bagid, sectionid)
										elseif IsAltKeyDown() then
											self:MoveSection(bagid, sectionid, true)
										elseif IsControlKeyDown() then 
											self:StartDeleteSection(bagid,sectionid) 
										else 
											currentsection = sectionid
										end 
										tablet:Refresh("BagginsEditBags") 
									end,
									"arg1",bagid,"arg2",sectionid)
				else
					cat:AddLine("text","- "..section.name,"size",12,"indentation", 20,"textR",1,"textG",1,"textB",1,
									"func",function(bagid,sectionid) 
										if IsShiftKeyDown() then
											self:MoveSection(bagid, sectionid)
										elseif IsAltKeyDown() then
											self:MoveSection(bagid, sectionid, true)
										elseif IsControlKeyDown() then 
											self:StartDeleteSection(bagid,sectionid) 
										else 
											currentsection = nil
										end 
										tablet:Refresh("BagginsEditBags") 
									end,
									"arg1",bagid,"arg2",sectionid)
									
					cat:AddLine("text",L["Categories"],"size",10,"indentation",30)
					for categoryid, category in pairs(section.cats) do
						cat:AddLine("text",category,"size",10,"indentation",40,"textR",1,"textG",1,"textB",1,
							"func",function(bagid,sectionid) 
									if IsControlKeyDown() then
										self:StartDeleteCategory(bagid,sectionid,categoryid) 
									end
							end,
							"arg1",bagid,"arg2",sectionid)
					end		
					--cat:AddLine("text"," ","size",10)
					cat:AddLine("text",L["Add Category"],"size",10,"indentation", 40,"textR",0.2,"textG",1,"textB",0.5,
							"func",function(bagid,sectionid) 
									dewdrop:Open(self.dewdropparent, "children", function()
																						while #catsorttable > 0 do
																							tremove(catsorttable,#catsorttable)
																						end
																						for catid in pairs(self.db.profile.categories) do
																							tinsert(catsorttable,catid)
																						end
																						table.sort(catsorttable)

																						for k, catid in ipairs(catsorttable) do
																						dewdrop:AddLine("text",catid,"func",function(category) self:AddCategory(bagid,sectionid,category) tablet:Refresh("BagginsEditBags") end,"arg1",catid,"closeWhenClicked",true)
																					end
																				end,
																	'point', "TOPLEFT",
																	'relativePoint', "TOPLEFT",
																	'cursorX', true, "cursorY", true)
							end,
							"arg1",bagid,"arg2",sectionid)
					cat:AddLine("text"," ","size",10)
				end

			end
			--cat:AddLine("text"," ","size",12)
			cat:AddLine("text",L["New Section"],"size",12,"indentation", 20,"textR",0.2,"textG",1,"textB",0.5,
							"func",function(bagid,sectionid) 
									StaticPopup_Show("BAGGINS_NEW_SECTION") 
							end,
							"arg1",bagid,"arg2",sectionid)
		end
	end
	cat = tablet:AddCategory("text",L["New Bag"],"size",14,"textR",0.2,"textG",1,"textB",0.5,"showWithoutChildren",true,"func", function() StaticPopup_Show("BAGGINS_NEW_BAG") end)
	cat:AddLine("text"," ","size",12)
	--ugh calling Close directly here breaks things, ugly hack!
	cat:AddLine("text",L["Close"],"func", function() self:ScheduleEvent(function() tablet:Close("BagginsEditBags") end, 0.01) end)
	tablet:SetHint(L["Click on an entry to open. Shift-Click to move up. Alt-Click to move down. Ctrl-Click to delete."])
end

function Baggins:FillCategoryEditTablet()
	local categories = self.db.profile.categories
	local cat
	tablet:SetTitle("Categories")
	tablet:SetTitleColor(1,1,1)
	
	while #catsorttable > 0 do
		tremove(catsorttable,#catsorttable)
	end
	for catid in pairs(categories) do
		tinsert(catsorttable,catid)
	end
	table.sort(catsorttable)
	
	for k, catid in ipairs(catsorttable) do
		local category = categories[catid]
		if currentcategory ~= catid then
			cat = tablet:AddCategory("id",catid,"text","+ "..catid,"size",14,"showWithoutChildren",true,"hideBlankLine",true,
										"func",function(catid) 
											if IsControlKeyDown() then 
												self:StartDeleteCategory(catid) 
											else 
												currentcategory = catid
											end 
											tablet:Refresh("BagginsEditCategories") 
										end,
										"arg1",catid)
		else
			cat = tablet:AddCategory("id",catid,"text","- "..catid,"size",14,"showWithoutChildren",true,"hideBlankLine",true,
										"func",function(catid) 
											if IsControlKeyDown() then 
												self:StartDeleteCategory(catid) 
											else 
												currentcategory = nil
											end 
											tablet:Refresh("BagginsEditCategories") 
										end,
										"arg1",catid)
			cat:AddLine("text",L["Rules"],"size",12,"indentation",10)
			for ruleid, rule in ipairs(category) do
				local rulename
				if rule.type == "New" then
					rulename = L["New Rule"]
				else
					rulename = RuleTypes[rule.type].GetName(rule)
				end
				if ruleid > 1 then
					rulename = (rule.operation or "OR").." "..rulename
				end
				cat:AddLine("text",rulename,"size",10,"indentation",20,"textR",1,"textG",1,"textB",1,
					"func",function(catid,ruleid,rule) 
							if IsShiftKeyDown() then
								self:MoveRule(catid, ruleid)
							elseif IsAltKeyDown() then
								self:MoveRule(catid, ruleid, true)
							elseif IsControlKeyDown() then
								self:StartDeleteRule(catid,ruleid) 
							else
								dewdrop:Open(self.dewdropparent, "children", function(...) Baggins:CreateRulesDewdrop(ruleid, rule, ...) end,
															'point', "TOPLEFT",
															'relativePoint', "TOPLEFT",
															'cursorX', true, "cursorY", true)
							end
					end,
					"arg1",catid,"arg2",ruleid,"arg3",rule)
			end		
			cat:AddLine("text",L["Add Rule"],"size",10,"indentation", 20,"textR",0.2,"textG",1,"textB",0.5,
					"func",function() 
							tinsert(category, { type="New" })
					end,
					"arg1",bagid,"arg2",sectionid)
			cat:AddLine("text"," ","size",10)
		end
	end
	cat = tablet:AddCategory("text",L["New Category"],"size",14,"textR",0.2,"textG",1,"textB",0.5,"showWithoutChildren",true,"func", function() StaticPopup_Show("BAGGINS_NEW_CATEGORY") end)
	cat:AddLine("text"," ","size",12)
	cat:AddLine("text",L["Apply"],"func", function() self:ForceFullUpdate() end)
	--ugh calling Close directly here breaks things, ugly hack!
	cat:AddLine("text",L["Close"],"func", function() self:ScheduleEvent(function() self:ForceFullUpdate() tablet:Close("BagginsEditCategories") self.CatEditOpen = false end, 0.01) end)
	tablet:SetHint(L["Click on an entry to open. Ctrl-Click to delete."])
end

function Baggins:setRuleOperation(rule,operation)
	rule.operation = operation
	self:OnRuleChanged()
end

function Baggins:setRuleType(rule,type)
	rule.type = type
	self:CleanRule(rule)
	self:OnRuleChanged()
end

function Baggins:CreateRulesDewdrop(ruleindex, rule, level, value, valueN_1, valueN_2, valueN_3, valueN_4)
	if level == 1 then
		dewdrop:AddLine('text', L["Editing Rule"], 'isTitle', true)
		
		dewdrop:AddLine('text', L["Type"], 'hasArrow', true,'value',"Type")
		dewdrop:AddLine()
		if not rule.type or rule.type == "New" then
			dewdrop:AddLine('text', L["Select a rule type to create the rule"], "isTitle",true)
		else
			RuleTypes[rule.type].DewDropOptions(rule, level, value, valueN_1, valueN_2, valueN_3, valueN_4)
			if ruleindex > 1 then
				dewdrop:AddLine('text', L["Operation"], 'hasArrow', true,'value',"Operation")
			end
		end
	elseif level == 2 and value == "Type" then
		for ruletype, ruledef in pairs(RuleTypes) do
			dewdrop:AddLine('text', ruledef.DisplayName, "checked", rule.type == ruletype,
				"tooltipTitle", ruledef.DisplayName, "tooltipText", ruledef.Description,
				"func",self.setRuleType, "arg1", self, "arg2", rule, "arg3", ruletype)
		end
	elseif level == 2 and value == "Operation" then
		dewdrop:AddLine('text', L["AND"], "checked", rule.operation == "AND","func",self.setRuleOperation, "arg1", self, "arg2", rule, "arg3", "AND")
		dewdrop:AddLine('text', L["OR"], "checked", rule.operation == "OR" or rule.operation == nil,"func",self.setRuleOperation, "arg1", self, "arg2", rule, "arg3", "OR")
		dewdrop:AddLine('text', L["NOT"], "checked", rule.operation == "NOT","func",self.setRuleOperation, "arg1", self, "arg2", rule, "arg3", "NOT")
	else
		RuleTypes[rule.type].DewDropOptions(rule,level,value, valueN_1, valueN_2, valueN_3, valueN_4)
	end
end

function Baggins:IsCategoryEditOpen()
	return self.CatEditOpen
end


---------------------
-- Fubar Plugin    --
---------------------
function Baggins:OnClick()
	if IsShiftKeyDown() then
		self:SaveItemCounts() 
		self:UpdateItemButtons()
	else
		self:ToggleAllBags()
	end
end

function Baggins:OnTooltipUpdate()
	local cat = tablet:AddCategory() 
	
	for bagid, bag in ipairs(self.db.profile.bags) do
		if not bag.isBank or (bag.isBank and self.bankIsOpen) then
			local color
			if bag.isBank then
				color = self.colors.blue
			else
				color = self.colors.white
			end
			local name = bag.name
			if self.bagframes[bagid] and self.bagframes[bagid]:IsVisible() then
				name = "* "..name
			end
			cat:AddLine("text",name,"func", function(self, bagid) 
													if IsShiftKeyDown() then
														self:MoveBag(bagid)
													elseif IsAltKeyDown() then
														self:MoveBag(bagid, true)
													else
														self:ToggleBag(bagid)
													end
												end ,"arg1",self,"arg2",bagid,
												"textR",color.r,"textG",color.g,"textB",color.b)
		end
	end
	tablet:SetHint(L["Click a bag to toggle it. Shift-click to move it up. Alt-click to move it down"])
end

function Baggins:BuildCountString(empty, total, color)
	local p = self.db.profile
	local s
	local divcolor 
	if self:IsTextColored() then
		divcolor = "|r"
	else
		divcolor = ""
	end
	if p.showused then
		s = color..total-empty
	end
	if p.showempty then
		if s then
			s = s..divcolor.."/"..color..empty
		else
			s = color..empty
		end
	end
	if p.showtotal then
		if s then
			s = s..divcolor.."/"..color..total
		else
			s = color..total
		end
	end
	return s
end

function Baggins:OnTextUpdate()
	local p = self.db.profile
	local text, color
	local allempty, alltotal = self:CountEmptySlots("ALL"), self:CountSlots("ALL")
	local normalempty, normaltotal = self:CountEmptySlots(), self:CountSlots()
	local soulempty, soultotal = self:CountEmptySlots("s"), self:CountSlots("s")
	local ammoempty, ammototal = self:CountEmptySlots("a"), self:CountSlots("a")
	
	local specialempty = allempty - normalempty - ammoempty - soulempty
	local specialtotal =  alltotal - normaltotal - ammototal - soultotal
	

	if p.combinecounts then
		local empty, total = normalempty, normaltotal
		if p.showammocount then
			empty = empty + ammoempty
			total = total + ammototal
		end
		if p.showsoulcount then
			empty = empty + soulempty
			total = total + soultotal
		end
		if p.showspecialcount then
			empty = empty + specialempty
			total = total + specialtotal
		end
		
		if self:IsTextColored() then
			local fullness = 1 - (empty/total)
			local r, g
			r = math.min(1,fullness * 2)
			g = math.min(1,(1-fullness) *2)
			color = ("|cFF%2X%2X00"):format(r*255,g*255)
		else
			color = ""
		end
		
		text = self:BuildCountString(empty,total,color)
		

	else
		local soulcount, ammocount, specialcount
		
		if self:IsTextColored() then
			local fullness = 1 - (normalempty/normaltotal)
			local r, g
			r = math.min(1,fullness * 2)
			g = math.min(1,(1-fullness) *2)
			color = ("|cFF%2X%2X00"):format(r*255,g*255)
		else
			color = ""
		end	
		
		text = self:BuildCountString(normalempty,normaltotal,color)
		
		if self.db.profile.showsoulcount and soultotal > 0 then
			if self:IsTextColored() then
				color = self.colors.purple.hex
			end			
			soulcount = self:BuildCountString(soulempty,soultotal,color)
			if soulcount then
				if text then
					text = text.." "..soulcount
				else
					text = soulcount
				end
			end
		end
		
		if self.db.profile.showammocount and ammototal > 0 then
			if self:IsTextColored() then
				color = self.colors.white.hex
			end
			ammocount = self:BuildCountString(ammoempty,ammototal,color)
			if ammocount then
				if text then
					text = text.." "..ammocount
				else
					text = ammocount
				end
			end
		end
		
		if self.db.profile.showspecialcount and specialtotal > 0 then
			if self:IsTextColored() then
				color = self.colors.blue.hex
			end
			specialcount = self:BuildCountString(specialempty,specialtotal,color)
			if specialcount then
				if text then
					text = text.." "..specialcount
				else
					text = specialcount
				end
			end
		end		
	end
			
	if text then
		self:SetText(text)	
	else
		self:SetText("Baggins")
	end
end


---------------------
-- Popup Dialogs   --
---------------------
local deletingbag
local deletingsection
local deletingcategory
local deletingrule
function Baggins:StartDeleteBag(bagid)
	deletingbag = bagid
	StaticPopup_Show("BAGGINS_DEL_BAG")
end

function Baggins:ConfirmDeleteBag()
	self:RemoveBag(deletingbag)
	deletingbag = nil
end

function Baggins:StartDeleteSection(bagid,sectionid)
	deletingbag = bagid
	deletingsection = sectionid
	StaticPopup_Show("BAGGINS_DEL_SECTION")
end

function Baggins:ConfirmDeleteSection()
	self:RemoveSection(deletingbag,deletingsection)
	deletingbag = nil
	deletingsection = nil
end

function Baggins:StartDeleteCategory(bagid,sectionid,categoryid)
	deletingbag = bagid
	deletingsection = sectionid
	deletingcategory = categoryid
	StaticPopup_Show("BAGGINS_DEL_CATEGORY")
end

function Baggins:ConfirmDeleteCategory()
	self:RemoveCategory(deletingbag,deletingsection,deletingcategory)
	deletingbag = nil
	deletingsection = nil
	deletingcategory = nil
end

function Baggins:StartDeleteRule(catid,ruleid)
	deletingcat = catid
	deletingrule = ruleid
	StaticPopup_Show("BAGGINS_DEL_RULE")
end

function Baggins:ConfirmDeleteRule()
	self:RemoveRule(deletingcat,deletingrule)
	deletingcat = nil
	deletingrule = nil
end

StaticPopupDialogs["BAGGINS_NEW_BAG"] = {
	text = L["Baggins - New Bag"],
	button1 = L["Accept"],
	button2 = L["Cancel"],
	hasEditBox = 1,
	maxLetters = 20,
	OnAccept = function()
		local editBox = getglobal(this:GetParent():GetName().."EditBox")
		Baggins:NewBag(editBox:GetText())
		tablet:Refresh("BagginsEditBags")
	end,
	OnShow = function()
		getglobal(this:GetName().."EditBox"):SetFocus()
	end,
	OnHide = function()
		if ( ChatFrameEditBox:IsVisible() ) then
			ChatFrameEditBox:SetFocus()
		end
		getglobal(this:GetName().."EditBox"):SetText("")
	end,
	EditBoxOnEnterPressed = function()
		local editBox = getglobal(this:GetParent():GetName().."EditBox")
		Baggins:NewBag(editBox:GetText())
		tablet:Refresh("BagginsEditBags")
		this:GetParent():Hide()
	end,
	EditBoxOnEscapePressed = function()
		this:GetParent():Hide()
	end,
	timeout = 0,
	exclusive = 1,
	whileDead = 1,
	hideOnEscape = 1
}
StaticPopupDialogs["BAGGINS_NEW_SECTION"] = {
	text = L["Baggins - New Section"],
	button1 = L["Accept"],
	button2 = L["Cancel"],
	hasEditBox = 1,
	maxLetters = 20,
	OnAccept = function()
		local editBox = getglobal(this:GetParent():GetName().."EditBox")
		Baggins:NewSection(currentbag,editBox:GetText())
		tablet:Refresh("BagginsEditBags")
	end,
	OnShow = function()
		getglobal(this:GetName().."EditBox"):SetFocus()
	end,
	OnHide = function()
		if ( ChatFrameEditBox:IsVisible() ) then
			ChatFrameEditBox:SetFocus()
		end
		getglobal(this:GetName().."EditBox"):SetText("")
	end,
	EditBoxOnEnterPressed = function()
		local editBox = getglobal(this:GetParent():GetName().."EditBox")
		Baggins:NewSection(currentbag,editBox:GetText())
		tablet:Refresh("BagginsEditBags")
		this:GetParent():Hide()
	end,
	EditBoxOnEscapePressed = function()
		this:GetParent():Hide()
	end,
	timeout = 0,
	exclusive = 1,
	whileDead = 1,
	hideOnEscape = 1
}
StaticPopupDialogs["BAGGINS_NEW_CATEGORY"] = {
	text = L["Baggins - New Category"],
	button1 = L["Accept"],
	button2 = L["Cancel"],
	hasEditBox = 1,
	maxLetters = 20,
	OnAccept = function()
		local editBox = getglobal(this:GetParent():GetName().."EditBox")
		Baggins:NewCategory(editBox:GetText())
		tablet:Refresh("BagginsEditCategories")
	end,
	OnShow = function()
		getglobal(this:GetName().."EditBox"):SetFocus()
	end,
	OnHide = function()
		if ( ChatFrameEditBox:IsVisible() ) then
			ChatFrameEditBox:SetFocus()
		end
		getglobal(this:GetName().."EditBox"):SetText("")
	end,
	EditBoxOnEnterPressed = function()
		local editBox = getglobal(this:GetParent():GetName().."EditBox")
		local name = editBox:GetText()
		this:GetParent():Hide()
		Baggins:NewCategory(name)
		tablet:Refresh("BagginsEditCategories")
		
	end,
	EditBoxOnEscapePressed = function()
		this:GetParent():Hide()
	end,
	timeout = 0,
	exclusive = 1,
	whileDead = 1,
	hideOnEscape = 1
}
StaticPopupDialogs["BAGGINS_DEL_BAG"] = {
	text = L["Are you sure you want to delete this Bag? this cannot be undone"],
	button1 = L["Delete"],
	button2 = L["Cancel"],
	OnAccept = function()
		Baggins:ConfirmDeleteBag()
		tablet:Refresh("BagginsEditBags")
	end,
	timeout = 0,
	exclusive = 1,
	hideOnEscape = 1
}
StaticPopupDialogs["BAGGINS_DEL_SECTION"] = {
	text = L["Are you sure you want to delete this Section? this cannot be undone"],
	button1 = L["Delete"],
	button2 = L["Cancel"],
	OnAccept = function()
		Baggins:ConfirmDeleteSection()
		tablet:Refresh("BagginsEditBags")
	end,
	timeout = 0,
	exclusive = 1,
	hideOnEscape = 1
}
StaticPopupDialogs["BAGGINS_DEL_CATEGORY"] = {
	text = L["Are you sure you want to remove this Category? this cannot be undone"],
	button1 = L["Delete"],
	button2 = L["Cancel"],
	OnAccept = function()
		Baggins:ConfirmDeleteCategory()
		tablet:Refresh("BagginsEditBags")
		tablet:Refresh("BagginsEditCategories")
	end,
	timeout = 0,
	exclusive = 1,
	hideOnEscape = 1
}
StaticPopupDialogs["BAGGINS_DEL_RULE"] = {
	text = L["Are you sure you want to remove this Rule? this cannot be undone"],
	button1 = L["Delete"],
	button2 = L["Cancel"],
	OnAccept = function()
		Baggins:ConfirmDeleteRule()
		tablet:Refresh("BagginsEditCategories")
	end,
	timeout = 0,
	exclusive = 1,
	hideOnEscape = 1
}
StaticPopupDialogs["BAGGINS_CATINUSE"] = {
	text = L["That category is in use by one or more bags, you cannot delete it."],
	button1 = "Ok",
	timeout = 0,
	exclusive = 1,
	hideOnEscape = 1
}
StaticPopupDialogs["BAGGINS_CATEXISTS"] = {
	text = L["A category with that name already exists."],
	button1 = "Ok",
	timeout = 0,
	exclusive = 1,
	hideOnEscape = 1
}

---------------------
-- Add/Remove/Move --
---------------------
function Baggins:NewBag(bagname)
	local bags = self.db.profile.bags
	tinsert(bags, { name=bagname, openWithAll=true, sections={}})
	if not self.bagframes[#bags] then
		self.BagFramePool:Acquire(#bags)
	end
	currentbag = #bags
	self:UpdateBags()
	if self.db.profile.layout == "auto" then
		self:LayoutBagFrames()
	end
end

function Baggins:MoveBag(bagid, down)
	local bags = self.db.profile.bags
	local other
	if down then
		other = bagid + 1
	else
		other = bagid - 1
	end
	
	if bags[bagid] and bags[other] then
		bags[bagid], bags[other] = bags[other], bags[bagid]
	end
end

function Baggins:MoveSection(bagid, sectionid, down)
	local other
	if down then
		other = sectionid + 1
	else
		other = sectionid - 1
	end
	if self.db.profile.bags[bagid] then
		local sections = self.db.profile.bags[bagid].sections
		if sections[sectionid] and sections[other] then
			sections[sectionid], sections[other] = sections[other], sections[sectionid]
		end
	end
end

function Baggins:MoveRule(categoryid, ruleid, down)
	local other
	if down then
		other = ruleid + 1
	else
		other = ruleid - 1
	end
	local rules = self.db.profile.categories[categoryid]
	if rules then
		if rules[ruleid] and rules[other] then
			rules[ruleid], rules[other] = rules[other], rules[ruleid]
		end
	end
end

function Baggins:RemoveBag(bagid)
	self:CloseBag(bagid)
	tremove(self.db.profile.bags, bagid)
	self:UpdateBags()
	if self.db.profile.layout == "auto" then
		self:LayoutBagFrames()
	end	
	local numbags, numbagframes = #self.db.profile.bags, #self.bagframes
	for i = numbags+1, numbagframes do
		if self.bagframes[i] then
			self.bagframes[i]:ReleastToPool();
		end
	end
end

function Baggins:NewSection(bagid,sectionname)
	tinsert(self.db.profile.bags[bagid].sections, { name=sectionname, cats = {} })
	currentsection = #self.db.profile.bags[bagid].sections
	self:UpdateBags()
	if self.db.profile.layout == "auto" then
		self:LayoutBagFrames()
	end	
end

function Baggins:NewCategory(catname)
	if not self.db.profile.categories[catname] then
		self.db.profile.categories[catname] = { name=catname } 
		currentcategory = catname
		tablet:Refresh("BagginsEditCategories") 
	else
		StaticPopup_Show("BAGGINS_CATEXISTS")
	end
end

function Baggins:RemoveSection(bagid, sectionid)
	tremove(self.db.profile.bags[bagid].sections, sectionid)
	self:UpdateBags()
	if self.db.profile.layout == "auto" then
		self:LayoutBagFrames()
	end	
end

function Baggins:RemoveRule(catid, ruleid)
	tremove(self.db.profile.categories[catid], ruleid)
	tablet:Refresh("BagginsEditCategories") 
	self:OnRuleChanged()
end

function Baggins:AddCategory(bagid,sectionid,category)
	tinsert(self.db.profile.bags[bagid].sections[sectionid].cats,category)
	self:UpdateBags()
	if self.db.profile.layout == "auto" then
		self:LayoutBagFrames()
	end	
end

function Baggins:CategoryInUse(catname)
	local inUse = false
	for bid, bag in ipairs(self.db.profile.bags) do
		for sid, section in ipairs(bag.sections) do
			for cid, cat in ipairs(section.cats) do
				if cat == catname then
					inUse = true
				end
			end
		end
	end
	return inUse
end

function Baggins:RemoveCategory(bagid,sectionid,catid)
	local p = self.db.profile
	if type(bagid) == "string" then
		--removing a category completely
		
		if not self:CategoryInUse(bagid) then
			p.categories[bagid] = nil
			tablet:Refresh("BagginsEditCategories") 
		else
			StaticPopup_Show("BAGGINS_CATINUSE")
		end	
	else
		--removing a category from a bag
		tremove(p.bags[bagid].sections[sectionid].cats,catid)
		self:UpdateBags()
		if p.layout == "auto" then
			self:LayoutBagFrames()	
		end	
	end
end

---------------------
-- Bag Open/Close  --
---------------------
function Baggins:CloseBag(bagid)
	if self.bagframes[bagid] then
		self.bagframes[bagid]:Hide()
		if self.db.profile.hidedefaultbank and self.db.profile.bags[bagid].isBank then
			if self.bankIsOpen and not self:IsAnyBankOpen() then
				CloseBankFrame()
			end
		end
		if self.db.profile.layout == "auto" then
			self:LayoutBagFrames()
		end
	end
	self:UpdateTooltip()
end

function Baggins:CloseAllBags()
	for bagid, bag in ipairs(self.db.profile.bags) do
		Baggins:CloseBag(bagid)	
	end
end

function Baggins:ToggleBag(bagid)
	if not self.bagframes[bagid] then
		self.BagFramePool:Acquire(bagid)
		self:OpenBag(bagid)
	elseif self.bagframes[bagid]:IsVisible() then
		self:CloseBag(bagid)
	else
		self:OpenBag(bagid)
	end
end

function Baggins:OpenBag(bagid,noupdate)
	local p = self.db.profile
	if not self:IsActive() then
		return
	end
	if self.doInitialUpdate then
		self:ForceFullUpdate()
		self.doInitialUpdate = nil
	end
	if p.bags[bagid].isBank and not self.bankIsOpen then
		return
	end
	if not self.bagframes[bagid] then
		self:CreateBagFrame(bagid)
	end
	self.bagframes[bagid]:Show()
	if not noupdate then
		self:UpdateBags()
	end
	if p.layout == "auto" then
		self:LayoutBagFrames()
	else
		self:LoadBagPosition(bagid)
	end
	self:UpdateTooltip()
end

-- "All Bags" in these 3 functions refers to bags that are set to openWithAll
function Baggins:OpenAllBags()
	local p = self.db.profile
	if not self:IsActive() then
		return
	end
	for bagid, bag in ipairs(p.bags) do
		if bag.openWithAll then
			Baggins:OpenBag(bagid,true)
		end			
	end
	self:UpdateBags()
	if p.layout == "auto" then
		self:LayoutBagFrames()
	else
		self:LoadBagPosition(bagid)
	end	
end

function Baggins:ToggleAllBags()
	if self:IsAllOpen() then
		self:CloseAllBags()
	else
		self:OpenAllBags()
	end
end

function Baggins:IsAllOpen()
	for bagid, bag in ipairs(self.db.profile.bags) do
		if bag.openWithAll and (not bag.isBank or self.bankIsOpen) then
			if not ( self.bagframes[bagid] and self.bagframes[bagid]:IsVisible() ) then
				return false
			end
		end			
	end
	return true
end

function Baggins:IsAnyOpen()
	for bagid, bag in ipairs(self.db.profile.bags) do
		if bag.openWithAll and (not bag.isBank or self.bankIsOpen) then
			if self.bagframes[bagid] and self.bagframes[bagid]:IsVisible() then
				return true
			end
		end			
	end
end

function Baggins:IsAnyBankOpen()
	for bagid, bag in ipairs(self.db.profile.bags) do
		if bag.isBank and self.bagframes[bagid] and self.bagframes[bagid]:IsVisible() then
			return true
		end			
	end
end

function Baggins:IsEmpty(bagid)
	local count = 0
	if self.bagframes[bagid] then
		for i, v in ipairs(self.bagframes[bagid].sections) do
			count = count + v.itemcount or 0
		end
	end
	return count == 0
end

function Baggins:OpenBackpack()
	self:OpenAllBags()
end

function Baggins:CloseBackpack()
	self:CloseAllBags()
end

function Baggins:CloseSpecialWindows()
    if self:IsAnyOpen() then
		self:CloseAllBags()
		return true
	end
	return self.hooks.CloseSpecialWindows()
end

---------------------
-- PT Set Browser  --
---------------------
function Baggins:BuildPTSetTable()
	ptsets = { }
	for modulename, moduletable in pairs(pt.k) do
		ptsets[modulename] = { }
		for setname, _ in pairs(moduletable) do
			if string.match(setname, " %- ") then
				local prefix, suffix = string.match(setname, "(.+) %- (.+)")
				if not ptsets[modulename][prefix] then
					tinsert(ptsets[modulename], prefix)
					ptsets[modulename][prefix] = {}
				end
				ptsets[modulename][prefix][suffix] = setname
				tinsert(ptsets[modulename][prefix], suffix)
			else
				if not ptsets[modulename][setname] then
					tinsert(ptsets[modulename], setname)
					ptsets[modulename][setname] = {}
				end
				ptsets[modulename][setname][setname] = setname
				tinsert(ptsets[modulename][setname], setname)
			end
		end
	end
	--sort the sets alphabetically
	for modname, module in pairs(ptsets) do
		table.sort(module)
		for setname, set in pairs(module) do
			if type(set) == "table" then
				table.sort(set)
			end
			--move the set to the top of the list of subsets
			for i = 1, #set do
				if set[i] == setname then
					tinsert(set,1,tremove(set,i))
					break
				end
			end
		end
	end
end

local ptsetstatus = {}
function Baggins:PTSetDewdrop(rule,level,value,value_1)
	local perlevel = 30
	
	if level == 1 then
		for k, v in pairs(ptsets) do
			dewdrop:AddLine("text",k,"hasArrow",true,"value",k)
		end
	elseif level == 2 then
		
		if ptsetstatus.value ~= value then
			ptsetstatus.start = 1
			ptsetstatus.value = value
		end
		if ptsetstatus.start ~= 1 then
			dewdrop:AddLine("text",L["Previous "]..perlevel,"func", function() ptsetstatus.start = ptsetstatus.start - perlevel end)
			dewdrop:AddLine()
		end
		for i = ptsetstatus.start, math.min(#ptsets[value], ptsetstatus.start + perlevel - 1) do
			local v = ptsets[value][i]
			dewdrop:AddLine("text",v,"hasArrow",true,"value", v)
		end
		if #ptsets[value] > ptsetstatus.start + perlevel then
			dewdrop:AddLine()
			dewdrop:AddLine("text",L["Next "]..perlevel,"func", function() ptsetstatus.start = ptsetstatus.start + perlevel end)
		end
		
		
	elseif level == 3 then
		if ptsetstatus.value2 ~= value_1..":"..value then
			ptsetstatus.start2 = 1
			ptsetstatus.value2 = value_1..":"..value
		end
		if ptsetstatus.start2 ~= 1 then
			dewdrop:AddLine("text",L["Previous "]..perlevel,"func", function() ptsetstatus.start2 = ptsetstatus.start2 - perlevel end)
			dewdrop:AddLine()
		end
		for i = ptsetstatus.start2, math.min(#ptsets[value_1][value], ptsetstatus.start2 + perlevel - 1) do
			local v = ptsets[value_1][value][i]
			local color
			if v == value then
				color = self.colors.blue
			else
				color = self.colors.white
			end
			dewdrop:AddLine("text",v,"hasArrow",false, "value", v, "func", function(name) rule.setname = name Baggins:OnRuleChanged() end,"arg1",ptsets[value_1][value][v],
				"textR",color.r,"textG",color.g,"textB",color.b,"closeWhenClicked",true)
		end
		if #ptsets[value_1][value] > ptsetstatus.start2 + perlevel then
			dewdrop:AddLine()
			dewdrop:AddLine("text",L["Next "]..perlevel,"func", function() ptsetstatus.start2 = ptsetstatus.start2 + perlevel end)
		end
	end
end

---------------------
--                 --
---------------------
